"" """
classifier-pipeline - this is a server side component that manipulates cptv
files and to create a classification model of animals present
Copyright (C) 2018, The Cacophony Project

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
"""

import os
import logging
import time
from multiprocessing import Process, Queue
import traceback
from ml_tools import tools

from track.clip import Clip, ClipStats
import numpy as np
import json
from pathlib import Path

import h5py
from cptv import CPTVReader
import yaml
from track.cliptrackextractor import is_affected_by_ffc
from track.track import Track
from track.region import Region

FPS = 9


def process_job(loader, queue, out_dir, config):
    i = 0
    while True:
        i += 1
        filename = queue.get()
        logging.info("Processing %s", filename)
        try:
            if filename == "DONE":
                break
            else:
                loader.process_file(str(filename), out_dir, config)
            if i % 50 == 0:
                logging.info("%s jobs left", queue.qsize())
        except Exception as e:
            logging.error("Process_job error %s %s", filename, e)
            traceback.print_exc()


class ClipLoader:
    def __init__(self, config):
        self.config = config

        # number of threads to use when processing jobs.
        self.workers_threads = 9

    def process_all(self, root, out_dir):
        out_dir.mkdir(parents=True, exist_ok=True)
        job_queue = Queue()
        processes = []
        for i in range(max(1, self.workers_threads)):
            p = Process(
                target=process_job,
                args=(self, job_queue, out_dir, self.config),
            )
            processes.append(p)
            p.start()

        file_paths = []
        #
        for folder_path, _, files in os.walk(root):
            for name in files:
                if os.path.splitext(name)[1] in [".cptv"]:
                    full_path = os.path.join(folder_path, name)
                    file_paths.append(full_path)
        # allows us know the order of processing
        file_paths.sort()
        for file_path in file_paths:
            job_queue.put(file_path)

        logging.info("Processing %d", job_queue.qsize())
        for i in range(len(processes)):
            job_queue.put("DONE")
        for process in processes:
            try:
                process.join()
            except KeyboardInterrupt:
                logging.info("KeyboardInterrupt, terminating.")
                for process in processes:
                    process.terminate()
                exit()

    def process_file(self, filename, out_dir, config):
        start = time.time()
        filename = Path(filename)
        logging.info(f"processing %s", filename)
        metadata_file = filename.with_suffix(".txt")
        if not metadata_file.exists():
            logging.error("No meta data found for %s", metadata_file)
            return

        metadata = tools.load_clip_metadata(metadata_file)
        r_id = metadata["id"]
        out_file = out_dir / f"{r_id}.hdf5"
        tracker_version = metadata.get("tracker_version")
        logging.info("Tracker version is %s", tracker_version)
        if out_file.exists() and tracker_version > 9:
            logging.warning("Already loaded %s", filename)
            # going to add some missing fierlds
            with h5py.File(out_file, "a") as f:
                if f.attrs.get("rec_time") is None:
                    with open(filename, "rb") as cptv:
                        reader = CPTVReader(cptv)
                        video_start_time = reader.timestamp.astimezone(Clip.local_tz)
                    f.attrs["rec_time"] = video_start_time.isoformat()
            return
        if len(metadata.get("Tracks")) == 0:
            logging.error("No tracks found for %s", filename)
            return

        clip = Clip(config.tracking["thermal"], filename)
        clip.load_metadata(
            metadata,
            config.build.tag_precedence,
        )

        with h5py.File(out_file, "w") as f:
            try:
                logging.info("creating clip %s", clip.get_id())

                clip_id = str(clip.get_id())

                clip_node = f
                triggered_temp_thresh = None
                camera_model = None
                with open(clip.source_file, "rb") as f:
                    reader = CPTVReader(f)
                    clip.set_res(reader.x_resolution, reader.y_resolution)
                    if clip.from_metadata:
                        for track in clip.tracks:
                            track.crop_regions()
                    if reader.model:
                        camera_model = reader.model.decode()
                    clip.set_model(camera_model)

                    # if we have the triggered motion threshold should use that
                    # maybe even override dynamic threshold with this value
                    if reader.motion_config:
                        motion = yaml.safe_load(reader.motion_config)
                        triggered_temp_thresh = motion.get("triggeredthresh")
                        if triggered_temp_thresh:
                            clip.temp_thresh = triggered_temp_thresh

                    video_start_time = reader.timestamp.astimezone(Clip.local_tz)
                    clip.set_video_stats(video_start_time)
                    print("Adding rec time", video_start_time)
                    frames = clip_node.create_group("frames")
                    ffc_frames = []
                    stats = ClipStats()
                    cropped_stats = ClipStats()
                    num_frames = 0
                    cptv_frames = []
                    region_adjust = 0
                    for frame in reader:
                        if frame.background_frame:
                            back_node = frames.create_dataset(
                                "background",
                                frame.pix.shape,
                                chunks=frame.pix.shape,
                                compression="gzip",
                                dtype=frame.pix.dtype,
                            )
                            back_node[:, :] = frame.pix
                            # pre tracker verison 10 there was a bug where back frame was counted in region frame number
                            #  so frame 0 is actually the background frame, no tracks shouild ever start at 0
                            if tracker_version < 10:
                                region_adjust = -1
                                logging.info("Adjusting regions by %s", region_adjust)
                            continue
                        ffc = is_affected_by_ffc(frame)
                        if ffc:
                            ffc_frames.append(num_frames)

                        cptv_frames.append(frame.pix)
                        stats.add_frame(frame.pix)
                        cropped_stats.add_frame(clip.crop_rectangle.subimage(frame.pix))
                        num_frames += 1
                    cptv_frames = np.uint16(cptv_frames)
                    thermal_node = frames.create_dataset(
                        "thermals",
                        cptv_frames.shape,
                        chunks=(1, *cptv_frames.shape[1:]),
                        compression="gzip",
                        dtype=cptv_frames.dtype,
                    )

                    thermal_node[:, :, :] = cptv_frames
                    stats.completed()
                    cropped_stats.completed()
                    group_attrs = clip_node.attrs
                    group_attrs["clip_id"] = r_id
                    group_attrs["rec_time"] = video_start_time.isoformat()
                    group_attrs["num_frames"] = np.uint16(num_frames)
                    group_attrs["ffc_frames"] = np.uint16(ffc_frames)
                    group_attrs["device_id"] = metadata["deviceId"]
                    stationId = metadata.get("stationId", 0)
                    group_attrs["station_id"] = stationId
                    group_attrs["crop_rectangle"] = np.uint8(
                        clip.crop_rectangle.to_ltrb()
                    )
                    group_attrs["max_temp"] = np.uint16(cropped_stats.max_temp)
                    group_attrs["min_temp"] = np.uint16(cropped_stats.min_temp)
                    group_attrs["mean_temp"] = np.uint16(cropped_stats.mean_temp)
                    group_attrs["frame_temp_min"] = np.uint16(
                        cropped_stats.frame_stats_min
                    )
                    group_attrs["frame_temp_max"] = np.uint16(
                        cropped_stats.frame_stats_max
                    )
                    group_attrs["frame_temp_median"] = np.uint16(
                        cropped_stats.frame_stats_median
                    )
                    group_attrs["frame_temp_mean"] = np.uint16(
                        cropped_stats.frame_stats_mean
                    )
                    group_attrs["start_time"] = clip.video_start_time.isoformat()
                    group_attrs["res_x"] = clip.res_x
                    group_attrs["res_y"] = clip.res_y
                    if camera_model is not None:
                        group_attrs["model"] = clip.camera_model

                    if triggered_temp_thresh is not None:
                        group_attrs["temp_thresh"] = triggered_temp_thresh

                    if clip.tags:
                        clip_tags = []
                        for track in clip.tags:
                            if track["what"]:
                                clip_tags.append(track["what"])
                            elif track["detail"]:
                                clip_tags.append(track["detail"])
                        group_attrs["tags"] = clip_tags

                tracks_group = clip_node.create_group("tracks")

                tracks = metadata.get("Tracks", [])
                for track in tracks:
                    track_id = track["id"]

                    track_group = tracks_group.create_group(str(track_id))

                    node_attrs = track_group.attrs
                    node_attrs["id"] = track_id
                    tags = track.get("tags", [])
                    tag = Track.get_best_human_tag(
                        tags, self.config.build.tag_precedence, 0
                    )

                    master_tag = [
                        t
                        for t in tags
                        if t.get("automatic")
                        and not isinstance(t.get("data", ""), str)
                        and t.get("data", {}).get("name") == "Master"
                    ]
                    if len(master_tag) > 0:
                        master_tag = master_tag[0]
                        node_attrs["ai_tag"] = master_tag["what"]
                        node_attrs["ai_tag_confidence"] = master_tag["confidence"]

                    if tag is not None:
                        node_attrs["human_tag"] = tag["what"]
                        node_attrs["human_tag_confidence"] = tag["confidence"]

                    human_tags = [
                        (t.get("what"), t["confidence"])
                        for t in tags
                        if t.get("automatic", False) != True
                    ]
                    if len(human_tags) > 0:
                        node_attrs["human_tags"] = [h[0] for h in human_tags]
                        node_attrs["human_tags_confidence"] = np.float32(
                            [h[1] for h in human_tags]
                        )

                    start = None
                    end = None

                    prev_frame = None
                    regions = []
                    for i, r in enumerate(track.get("positions")):
                        if isinstance(r, list):
                            region = Region.region_from_array(r[1])
                            if region.frame_number is None:
                                if i == 0:
                                    frame_number = round(r[0] * FPS)
                                    region.frame_number = frame_number
                                else:
                                    region.frame_number = prev_frame + 1
                        else:
                            region = Region.region_from_json(r)
                        if region.frame_number is None:
                            if "frameTime" in r:
                                if i == 0:
                                    region.frame_number = round(r["frameTime"] * 9)
                                else:
                                    region.frame_number = prev_frame + 1
                        # new_f = region.frame_number + region_adjust
                        prev_frame = region.frame_number
                        region.frame_number = region.frame_number + region_adjust
                        assert region.frame_number >= 0
                        regions.append(region.to_array())
                        if start is None:
                            start = region.frame_number
                        end = region.frame_number
                    node_attrs["start_frame"] = start
                    node_attrs["end_frame"] = min(num_frames, end)

                    region_array = np.uint16(regions)
                    regions = track_group.create_dataset(
                        "regions",
                        region_array.shape,
                        chunks=region_array.shape,
                        compression="gzip",
                        dtype=region_array.dtype,
                    )
                    regions[:, :] = region_array
            except:
                logging.error("Error saving file %s", filename, exc_info=True)
                f.close()
                out_file.unlink()



from pathlib import Path
import os.path as path

import attr
import logging
import yaml

from .trackingconfig import TrackingConfig
from .trainconfig import TrainConfig
from .classifyconfig import ClassifyConfig
from .buildconfig import BuildConfig
from .defaultconfig import DefaultConfig, deep_copy_map_if_key_not_exist

CONFIG_FILENAME = "classifier.yaml"
CONFIG_DIRS = [Path("/etc/cacophony"), Path(__file__).parent.parent]


@attr.s
class Config(DefaultConfig):
    DEFAULT_LABELS = [
        "bird",
        "cat",
        "false-positive",
        "hedgehog",
        "insect",
        "leporidae",
        "mustelid",
        "possum",
        "rodent",
        "wallaby",
    ]
    base_folder = attr.ib()
    labels = attr.ib()
    build = attr.ib()
    tracking = attr.ib()
    train = attr.ib()
    classify = attr.ib()
    reprocess = attr.ib()
    previews_colour_map = attr.ib()
    worker_threads = attr.ib()
    debug = attr.ib()
    use_opt_flow = attr.ib()
    verbose = attr.ib()

    @classmethod
    def load_from_file(cls, filename=None):
        if filename is None or not Path(filename).exists():
            filename = find_config()
        if filename is None:
            return Config.get_defaults()
        logging.info("Loading config from %s", filename)
        with open(filename) as stream:
            return cls.load_from_stream(stream)

    @classmethod
    def load_from_stream(cls, stream):
        raw = yaml.safe_load(stream)
        default = Config.get_defaults()
        if raw is None:
            raw = {}
        # Configuration from "tracking" section is used in
        # "classify_tracking" when not specified.
        deep_copy_map_if_key_not_exist(default.as_dict(), raw)
        base_folder = Path(raw.get("base_data_folder", "."))
        return cls(
            base_folder=Path(base_folder),
            tracking=TrackingConfig.load(raw["tracking"]),
            train=TrainConfig.load(raw["train"], base_folder),
            classify=ClassifyConfig.load(raw["classify"]),
            reprocess=raw["reprocess"],
            previews_colour_map=raw["previews_colour_map"],
            worker_threads=raw["worker_threads"],
            labels=raw["labels"],
            build=BuildConfig.load(raw["build"]),
            debug=raw["debug"],
            use_opt_flow=raw["use_opt_flow"],
            verbose=raw["verbose"],
        )

    @classmethod
    def get_defaults(cls):
        return cls(
            base_folder=".",
            labels=Config.DEFAULT_LABELS,
            reprocess=True,
            previews_colour_map="custom_colormap.dat",
            worker_threads=0,
            build=BuildConfig.get_defaults(),
            tracking=TrackingConfig.get_defaults(),
            train=TrainConfig.get_defaults(),
            classify=ClassifyConfig.get_defaults(),
            debug=False,
            use_opt_flow=False,
            verbose=False,
        )

    def validate(self):
        self.build.validate()
        for tracker in self.tracking.values():
            tracker.validate()
        self.train.validate()
        self.classify.validate()
        return True

    def as_dict(self):
        return attr.asdict(self)


def find_config():
    for directory in CONFIG_DIRS:
        p = directory / CONFIG_FILENAME
        logging.info("Looking for config %s", p)
        if p.is_file():
            return str(p)
    return None


def parse_options_param(name, value, options):
    if value is None:
        lower_value = value
    else:
        lower_value = value.lower()
    if lower_value not in options:
        raise Exception(
            "Cannot parse {} as '{}'.  Valid options are {}.".format(
                name, value, options
            )
        )
    return lower_value



#!/usr/bin/python3
import argparse
from datetime import datetime, timedelta
import logging
import os
import psutil
import socket
import time

import numpy as np
from threading import Thread

from config.timewindow import WindowStatus, TimeWindow, RelAbsTime
from config.config import Config
from config.thermalconfig import ThermalConfig
from .headerinfo import HeaderInfo
from ml_tools.logs import init_logging
from ml_tools.rectangle import Rectangle
from .piclassifier import PiClassifier, run_classifier
from .cameras import lepton3
import multiprocessing
from .eventreporter import log_event
from piclassifier.monitorconfig import monitor_file

SOCKET_NAME = "/var/run/lepton-frames"
VOSPI_DATA_SIZE = 160
TELEMETRY_PACKET_COUNT = 4
STOP_SIGNAL = "stop"

SKIP_SIGNAL = "skip"
SNAPSHOT_SIGNAL = "snap"

restart_pending = False
connected = False


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument("--file", help="a test file to send", default=None)
    parser.add_argument(
        "-p",
        "--preview-type",
        help="Create MP4 previews of this type",
    )
    parser.add_argument(
        "-c",
        "--config-file",
        help="Path to config file to use",
    )

    parser.add_argument(
        "--thermal-config-file", help="Path to pi-config file (config.toml) to use"
    )
    parser.add_argument(
        "--ir", action="count", help="Path to pi-config file (config.toml) to use"
    )

    parser.add_argument(
        "--fps",
        type=int,
        default=None,
        help="When running a file through specify the frame rate you want it to run at, otherwise it runs as fast as the cpu can",
    )

    args = parser.parse_args()
    return args


# Links to socket and continuously waits for 1 connection
def main():
    init_logging()
    args = parse_args()

    config = Config.load_from_file(args.config_file)
    thermal_config = ThermalConfig.load_from_file(args.thermal_config_file)
    monitor_thread = Thread(
        target=monitor_file, args=(file_changed, thermal_config.config_file)
    )
    monitor_thread.daemon = True
    monitor_thread.start()
    thermal_config.recorder.rec_window.set_location(
        *thermal_config.location.get_lat_long(use_default=True),
        thermal_config.location.altitude,
    )

    if args.file:
        return parse_file(
            args.file, config, thermal_config, args.preview_type, args.fps
        )

    process_queue = multiprocessing.Queue()

    # get a cloned window so we dont update it
    if not thermal_config.recorder.use_low_power_mode:
        snapshot_thread = Thread(
            target=take_snapshots,
            args=(
                thermal_config.recorder.rec_window.clone(),
                process_queue,
            ),
        )
        snapshot_thread.daemon = True
        snapshot_thread.start()
    if args.ir or thermal_config.device_setup.ir:
        while True:
            if restart_pending:
                break
            try:
                read = ir_camera(config, thermal_config, process_queue)
                if read == 0:
                    logging.error("Error reading camera try again in 10")
                    time.sleep(10)
                else:
                    log_event("camera-disconnected", f"read {read} frames")
            except Exception as ex:
                log_event("camera-disconnected", ex)
                logging.error("Error reading camera try again in 10", exc_info=True)
                time.sleep(10)
        return

    try:
        os.unlink(SOCKET_NAME)
    except OSError:
        if os.path.exists(SOCKET_NAME):
            raise
    logging.info("running as thermal")

    sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
    sock.bind(SOCKET_NAME)
    sock.settimeout(3 * 60)  # 3 minutes
    sock.listen(1)

    global connected
    while True:
        if restart_pending:
            sock.close()
            break
        logging.info("waiting for a connection")
        try:
            connection, client_address = sock.accept()
            connected = True
            logging.info("connection from %s", client_address)
            log_event("camera-connected", {"type": "thermal"})
            handle_connection(
                connection, config, args.thermal_config_file, process_queue
            )
        except socket.timeout:
            logging.error("Socket %s timeout error", SOCKET_NAME, exc_info=True)
            return

        except Exception as ex:
            log_event("camera-disconnected", ex)
            logging.error("Error with connection", exc_info=True)
        finally:
            # Clean up the connection
            try:
                connection.close()
            except:
                pass
        connected = False


def file_changed(event):
    logging.info("Received file changed event %s restarting", event)
    global restart_pending
    restart_pending = True
    if not connected:
        logging.info("Not conencted so closing")
        os._exit(0)


def parse_file(file, config, thermal_config, preview_type, fps):
    _, ext = os.path.splitext(file)
    thermal_config.recorder.rec_window = rec_window = TimeWindow(
        RelAbsTime(""), RelAbsTime(""), None, None, 0
    )

    if ext == ".cptv":
        parse_cptv(file, config, thermal_config.config_file, preview_type, fps)
    else:
        parse_ir(file, config, thermal_config, preview_type, fps)


def parse_ir(file, config, thermal_config, preview_type, fps):
    from piclassifier import irmotiondetector
    import cv2

    irmotiondetector.MIN_FRAMES = 0
    count = 0
    vidcap = cv2.VideoCapture(file)
    while True:
        if fps is not None:
            time.sleep(1 / fps)
        success, image = vidcap.read()
        if not success:
            break
        # gray = cv2.resize(gray, (640, 480))
        if count == 0:
            res_y, res_x = image.shape[:2]
            headers = HeaderInfo(
                res_x=res_x,
                res_y=res_y,
                fps=10,
                brand=None,
                model="IR",
                frame_size=res_y * res_x,
                pixel_bits=8,
                serial="",
                firmware="",
            )

            pi_classifier = PiClassifier(
                config,
                thermal_config,
                headers,
                thermal_config.motion.run_classifier,
                0,
                preview_type,
            )
            gray = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
            pi_classifier.motion_detector._background._background = np.float32(gray)
            pi_classifier.motion_detector._background.update_background(
                gray, learning_rate=1
            )
            pi_classifier.motion_detector._background._frames = 1000
            count += 1
            # assume this has been run over 1000 frames
            continue
        # frame = Frame(image, None, None, None, None)
        # frame.received_at = time.time()
        pi_classifier.process_frame(image, time.time())
        count += 1
    vidcap.release()
    pi_classifier.disconnected()


def preview_socket(headers, frame_queue):
    import yaml

    # convert casing
    python_dic = headers.__dict__
    go_dic = {}
    for k, v in python_dic.items():
        new_key = f"{k[0].upper()}{k[1:]}"
        try:
            under_index = new_key.index("_")
            new_key = f"{new_key[:under_index]}{new_key[under_index+1].upper()}{new_key[under_index+2:]}"
        except:
            pass
        go_dic[new_key] = v
    header_bytes = yaml.dump(go_dic).encode()
    header_bytes += b"\nclear"

    while True:
        try:
            # connect to management socket
            frameSocket = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
            frameSocket.connect("/var/spool/managementd")
            logging.info("Connected to management interface")
            frameSocket.send(header_bytes)
            telemetry_bytes = bytearray(640)
            # if we need this can add the correct info
            while True:
                frame = frame_queue.get()
                if isinstance(frame, str):
                    if frame == STOP_SIGNAL:
                        return
                if frame is None:
                    logging.info("Disconnected")
                    break
                frame_bytes = frame.pix.byteswap().tobytes()
                frame_bytes = telemetry_bytes + frame_bytes
                frameSocket.send(frame_bytes)
        except:
            # logging.error("Failed to connect to /var/spool/managementd", exc_info=True)
            try:
                # empty the queue
                items = frame_queue.qsize()
                items = max(items, 1)
                for _ in range(items):
                    item = frame_queue.get(100)
                    if isinstance(item, str):
                        if item == STOP_SIGNAL:
                            return
            except:
                pass
            # could not connect wait a few seconds
            time.sleep(2)


def parse_cptv(file, config, thermal_config_file, preview_type, fps):
    from cptv import Frame
    from cptv_rs_python_bindings import CptvReader

    reader = CptvReader(str(file))
    header = reader.get_header()
    telemetry_size = 160 * 4
    headers = HeaderInfo(
        res_x=header.x_resolution,
        res_y=header.y_resolution,
        fps=9,
        brand=header.brand if header.brand else None,
        model=header.model if header.model else None,
        frame_size=header.x_resolution * header.y_resolution * 2 + telemetry_size,
        pixel_bits=16,
        serial="",
        firmware="",
    )

    frame_queue = multiprocessing.Queue()
    preview_process = multiprocessing.Process(
        target=preview_socket,
        args=(
            headers,
            frame_queue,
        ),
    )
    preview_process.start()
    try:
        thermal_config = ThermalConfig.load_from_file(
            thermal_config_file, headers.model
        )

        pi_classifier = PiClassifier(
            config,
            thermal_config,
            headers,
            thermal_config.motion.run_classifier,
            0,
            preview_type,
        )
        while True:
            frame = reader.next_frame()

            if frame is None:
                break
            # to get extra properties and allow pickling convert to cptv.Frame
            frame = Frame(
                frame.pix,
                timedelta(milliseconds=frame.time_on),
                timedelta(milliseconds=frame.last_ffc_time),
                frame.temp_c,
                frame.last_ffc_temp_c,
                frame.background_frame,
            )

            frame_queue.put(frame)

            frame.ffc_imminent = False
            frame.ffc_status = 0

            if frame.background_frame:
                pi_classifier.motion_detector._background._background = frame.pix
                continue
            pi_classifier.process_frame(frame, time.time())
            if fps is not None:
                time.sleep(1.0 / fps)
        pi_classifier.disconnected()
        frame_queue.put(STOP_SIGNAL)
        preview_process.join(7)
        if preview_process.is_alive():
            logging.info("Killing preview process")
            try:
                preview_process.kill()
            except:
                pass
    except Exception as ex:
        pi_classifier.disconnected()
        logging.error("EXception all done")
        frame_queue.put(STOP_SIGNAL)
        preview_process.join(7)
        if preview_process.is_alive():
            logging.info("Killing preview process")
            try:
                preview_process.kill()
            except:
                pass
        raise ex


def get_processor(process_queue, config, thermal_config, headers):
    p_processor = multiprocessing.Process(
        target=run_classifier,
        args=(
            process_queue,
            config,
            thermal_config,
            headers,
            thermal_config.motion.run_classifier,
        ),
    )
    return p_processor


def handle_headers(connection):
    headers = b""
    left_over = None
    while True:
        logging.info("Getting header info")
        data = connection.recv(4096)

        if not data:
            raise Exception("Disconnected from camera while getting headers")
        headers += data
        done = headers.find(b"\n\n")
        if done > -1:
            left_over = headers[done + 2 :]
            headers = headers[:done]
            if left_over[:5] == b"clear":
                left_over = left_over[5:]
            break
    return HeaderInfo.parse_header(headers.decode()), left_over


def ir_camera(config, thermal_config, process_queue):
    import cv2

    FPS = 10
    logging.info("Starting ir video capture")
    cap = cv2.VideoCapture(0)
    cap.set(cv2.CAP_PROP_FPS, FPS)
    frames = 0
    try:
        res_x = cap.get(cv2.CAP_PROP_FRAME_WIDTH)
        res_y = cap.get(cv2.CAP_PROP_FRAME_HEIGHT)
        headers = HeaderInfo(
            res_x=int(res_x),
            res_y=int(res_y),
            fps=FPS,
            brand=None,
            model="IR",
            frame_size=res_y * res_x,
            pixel_bits=8,
            serial="",
            firmware="",
        )

        global connected
        connected = True
        processor = get_processor(process_queue, config, thermal_config, headers)
        processor.start()
        drop_frame = None
        dropped = 0
        start_dropping = None

        while True:
            if restart_pending:
                logging.info("Restarting as config changed")
                process_queue.put(STOP_SIGNAL)
                # give it time to clean up
                processor.join(5)
                if processor.is_alive():
                    logging.info("Killing process")
                    try:
                        processor.kill()
                    except:
                        pass
                break
            returned, frame = cap.read()
            if not processor.is_alive():
                logging.info("Processor stopped, restarting %s", processor.is_alive())
                processor = get_processor(
                    process_queue, config, thermal_config, headers
                )
                processor.start()
            if not returned:
                logging.info("no frame from video capture")
                process_queue.put(STOP_SIGNAL)
                break
            frames += 1
            if frames == 1:
                log_event("camera-connected", {"type": "IR"})
            if drop_frame is not None and (frames - start_dropping) % drop_frame == 0:
                logging.info("Dropping frame due to slow processing")
                dropped += 1
            else:
                process_queue.put((frame, time.time()))
            qsize = process_queue.qsize()
            if qsize > headers.fps * 4 and (
                drop_frame is None or frames > (start_dropping + drop_frame)
            ):
                # drop every 9th frame
                if drop_frame is None:
                    drop_frame = 9
                else:
                    drop_frame = drop_frame - 1
                # drop first frame
                start_dropping = frames + 1
                logging.info("Dropping every %s frame as qsize %s", drop_frame, qsize)
            elif qsize < headers.fps * 3:
                drop_frame = None
                start_dropping = None
    finally:
        if processor is not None:
            time.sleep(5)
            processor.kill()
    return frames


def next_snapshot(window, prev_window_type=None):
    current_status = None
    if prev_window_type is None:
        current_status = window.window_status()
    if window.non_stop:
        if prev_window_type is not None:
            window.next_window()
        return (window.start.dt, WindowStatus.non_stop)
    if current_status == WindowStatus.before or (
        prev_window_type == WindowStatus.after
    ):
        return (window.next_start(), WindowStatus.before)
    elif not window.non_stop and (
        current_status == WindowStatus.inside or prev_window_type == WindowStatus.before
    ):
        started = window.next_start()
        if (
            current_status is not None
            and abs((started - datetime.now()).total_seconds()) < 60 * 30
        ):
            logging.info("Started inside window within 30 mins")
            return (started, WindowStatus.before)

        return (window.next_end(), WindowStatus.inside)
    else:
        # next windowtimes
        window.next_window()
        return (window.next_start(), WindowStatus.before)


def take_snapshots(window, process_queue):
    if window.non_stop:
        window.start.dt = datetime.now()
        window.end.dt = datetime.now()
    next_snap = next_snapshot(window, None)
    while True:
        if next_snap is None:
            snap_time = datetime.now()
        else:
            snap_time = next_snap[0] - timedelta(minutes=2)
        time_until = (snap_time - datetime.now()).total_seconds()
        if time_until > 0:
            logging.info("Taking snapshot at %s", snap_time)
            time.sleep(time_until)
        logging.info("Sending snapshot signal")
        process_queue.put(SNAPSHOT_SIGNAL)
        next_snap = next_snapshot(window, next_snap[1])


def handle_connection(connection, config, thermal_config_file, process_queue):
    headers, extra_b = handle_headers(connection)
    thermal_config = ThermalConfig.load_from_file(thermal_config_file, headers.model)
    logging.info(
        "parsed camera headers %s running with config %s", headers, thermal_config
    )

    processor = get_processor(process_queue, config, thermal_config, headers)
    processor.start()

    edge = config.tracking["thermal"].edge_pixels
    crop_rectangle = Rectangle(
        edge, edge, headers.res_x - 2 * edge, headers.res_y - 2 * edge
    )
    raw_frame = lepton3.Lepton3(headers)
    read = 0
    try:
        while True:
            if restart_pending:
                logging.info("Restarting as config changed")
                process_queue.put(STOP_SIGNAL)
                # give it time to clean up
                processor.join(5)
                if processor.is_alive():
                    logging.info("Killing process")
                    try:
                        processor.kill()
                    except:
                        pass
                break
            if not processor.is_alive():
                # this potentially loops on indefinately on an error if the error is to do with the headers
                logging.info("Processor stopped restarting")
                processor = get_processor(
                    process_queue, config, thermal_config, headers
                )
                processor.start()
            if extra_b is not None:
                data = extra_b + connection.recv(
                    headers.frame_size - len(extra_b), socket.MSG_WAITALL
                )
                extra_b = None
            else:
                data = connection.recv(headers.frame_size, socket.MSG_WAITALL)

            if not data:
                logging.info("disconnected from camera")
                process_queue.put(STOP_SIGNAL)
                break
            try:
                message = data[:5]
                if message == b"clear":
                    logging.info(
                        "processing error from camera"
                    )  # TODO Check if this is handled properly.
                    process_queue.put(STOP_SIGNAL)
                    break
            except:
                pass
            read += 1

            frame = raw_frame.parse(data)
            frame.received_at = time.time()
            cropped_frame = crop_rectangle.subimage(frame.pix)
            t_min = np.amin(cropped_frame)
            # seems to happen if pi is working hard
            if t_min == 0:
                logging.warning(
                    "received frame has odd values skipping thermal frame min {} cpu % {} memory % {}".format(
                        t_min, psutil.cpu_percent(), psutil.virtual_memory()[2]
                    )
                )
                log_event("bad-thermal-frame", f"Bad Pixel of {t_min}")
                process_queue.put(SKIP_SIGNAL)
            else:
                process_queue.put((frame, time.time()))

    finally:
        time.sleep(5)
        # give it a moment to close down properly
        processor.terminate()



import numpy as np
import cv2
from .motiondetector import SlidingWindow, MotionDetector
from track.cliptracker import (
    CVBackground,
    Background,
    get_diff_back_filtered,
)


class RollingBackground(Background):
    AVERAGE_OVER = 1000

    def __init__(self, background_thresh=15):
        self._background = None
        self.frames = 0
        self.background_thresh = background_thresh

    def update_background(self, frame):
        if self._background is None:
            self._background = np.float32(frame.copy())
            return
        if self.frames < Background.AVERAGE_OVER:
            self._background = (self._background * self.frames + frame) / (
                self.frames + 1
            )
        else:
            self._background = (
                self._background * (Background.AVERAGE_OVER - 1) + frame
            ) / (Background.AVERAGE_OVER)

        self.frames += 1

    @property
    def background(self):
        return np.uint8(self._background)

    @property
    def frames(self):
        return min(self._frames, AVERAGE_OVER)

    def compute_filtered(self, thermal):
        return get_diff_back_filtered(
            self.background,
            thermal,
            self.background_thresh,
        )


WINDOW_SIZE = 50
MIN_FRAMES = 10 * 10  # 10 * 10  # 10seconds

THRESHOLD = 12
# maybe put to 12
TRIGGER_FRAMES = 2


class IRMotionDetector(MotionDetector):
    def __init__(self, thermal_config, headers):
        super().__init__(thermal_config, headers)
        self.num_preview_frames = thermal_config.recorder.preview_secs * headers.fps
        self.rgb_window = SlidingWindow(self.num_preview_frames, dtype=np.uint8)
        self.gray_window = SlidingWindow(self.num_preview_frames, dtype=np.uint8)
        # self._background = Background()
        self._background = CVBackground()
        self.kernel_trigger = np.ones(
            (15, 15), "uint8"
        )  # kernel for erosion when not recording
        self.kernel_recording = np.ones(
            (10, 10), "uint8"
        )  # kernel for erosion when recording
        self.movement_detected = False
        self.triggered = 0
        self.show = False
        self.prev_triggered = False

    def disconnected(self):
        self.rgb_window.reset()
        # self.gray_window.reset()
        self.processed = 0

    @property
    def calibrating(self):
        return False

    @property
    def background(self):
        return self._background.background

    def get_kernel(self):
        if self.movement_detected:
            return self.kernel_recording
        else:
            return self.kernel_trigger

    def preview_frames(self):
        return self.rgb_window.get_frames()[:-1]

    def get_recent_frame(self):
        return self.rgb_window.current

    # Processes a frame returning True if there is motion.
    def process_frame(self, frame, force_process=False):
        if self.can_record() or force_process:
            self.rgb_window.add(frame)
            gray = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)
            self.gray_window.add(gray)
            #
            if self.gray_window.oldest is None:
                return False
            learning_rate = 0 if self.movement_detected else -1
            self._background.update_background(gray, learning_rate=learning_rate)
            if self.num_frames > MIN_FRAMES:
                # Filter and get diff from background
                delta = cv2.absdiff(
                    self.gray_window.oldest, gray
                )  # Get delta from current frame and background

                threshold = cv2.threshold(delta, THRESHOLD, 255, cv2.THRESH_BINARY)[1]
                #
                erosion_image = cv2.erode(threshold, self.get_kernel())
                diff_erosion_pixels = len(erosion_image[erosion_image > 0])
                erosion_image = cv2.erode(
                    self._background.compute_filtered(None), self.get_kernel()
                )
                erosion_pixels = len(erosion_image[erosion_image > 0])

                # assert erosion_pixels == self._background.detect_motion()
                # if any have no pixels lets stop motion detection, wiht not updating
                # background when motion is detected if the background actually changes
                # this could cause a problem, this should catch that
                if self.movement_detected:
                    erosion_pixels = min(diff_erosion_pixels, erosion_pixels)
                # to do find a value that suites the number of pixesl we want to move
                # Calculate if there was motion in the current frame
                # TODO Chenage how much ioldests added to the triggered depending on how big the motion is
                self.prev_triggered = erosion_pixels > 0
                if erosion_pixels > 0:
                    self.triggered += 1
                    self.triggered = min(self.triggered, 30)
                else:
                    self.triggered -= 1
                    self.triggered = max(self.triggered, 0)
                # Check if motion has started or ended
                if not self.movement_detected and self.triggered >= TRIGGER_FRAMES:
                    self.movement_detected = True

                elif self.movement_detected and self.triggered <= 0:
                    self.movement_detected = False
        else:
            self.rgb_window.update_current_frame(frame)
        self.num_frames += 1
        return self.movement_detected

    @property
    def temp_thresh(self):
        return None



import attr


@attr.s
class LocationConfig:
    DEFAULT_LAT = -43.5321
    DEFAULT_LONG = 172.6362

    latitude = attr.ib()
    longitude = attr.ib()
    loc_timestamp = attr.ib()
    altitude = attr.ib()
    accuracy = attr.ib()

    @classmethod
    def load(cls, raw):
        return cls(
            latitude=raw.get("latitude", 0),
            longitude=raw.get("longitude", 0),
            loc_timestamp=raw.get("timestamp"),
            altitude=raw.get("altitude"),
            accuracy=raw.get("accuracy"),
        )

    def get_lat_long(self, use_default=False):
        lat = self.latitude
        lng = self.longitude
        if use_default and lat == 0:
            lat = LocationConfig.DEFAULT_LAT
        if use_default and lng == 0:
            lng = LocationConfig.DEFAULT_LONG
        return (lat, lng)



import json
import dateutil
import numpy as np
import logging
from ml_tools import tools
from track.region import Region
from abc import ABC, abstractmethod
from ml_tools.rectangle import Rectangle
from config.buildconfig import BuildConfig
from ml_tools import imageprocessing
from enum import Enum
import attr

FRAMES_PER_SECOND = 9

CPTV_FILE_WIDTH = 160
CPTV_FILE_HEIGHT = 120
FRAME_SIZE = 32
MIN_SIZE = 4

# hard coded for now
FP_LABELS = ["other", "unidentified", "rain", "false-positive", "water", "insect"]


class SegmentType(Enum):
    IMPORTANT_RANDOM = 0
    ALL_RANDOM = 1
    IMPORTANT_SEQUENTIAL = 2
    ALL_SEQUENTIAL = 3
    TOP_SEQUENTIAL = 4
    ALL_SECTIONS = 5
    TOP_RANDOM = 6
    ALL_RANDOM_NOMIN = 7
    ALL_RANDOM_MASKED = 8


class BaseSample(ABC):
    @property
    @abstractmethod
    def track_bounds(cls):
        """Get all regions for this sample"""
        ...

    @property
    @abstractmethod
    def source_file(self):
        """Source file."""
        ...

    @property
    @abstractmethod
    def track_median_mass(self):
        """The function gets median mass of the track."""
        ...

    @property
    @abstractmethod
    def frame_indices(self):
        """The function gets all frames indices for this sample."""
        ...

    @property
    @abstractmethod
    def unique_track_id(self):
        """Represent the unique identifier for this track."""
        ...

    @property
    @abstractmethod
    def unique_id(self):
        """Represent the unique identifier for this sample."""
        ...

    @property
    @abstractmethod
    def bin_id(self):
        """Represent the unique identifier for this sample."""
        ...

    @property
    @abstractmethod
    def sample_weight(self):
        """Represent the unique identifier for this sample."""
        ...

    @property
    @abstractmethod
    def mass(cls):
        """Get mass for this sample"""
        ...


class Sample(BaseSample):
    def __init__(self, label):
        self.original_label = label
        self.remapped_label = label

    @property
    def label(self):
        return self.remapped_label


EDGE = 1

res_x = 120
res_y = 160


@attr.s
class ClipHeader:
    clip_id = attr.ib()
    location = attr.ib()
    station_id = attr.ib()
    rec_time = attr.ib()
    source_file = attr.ib()
    frames_per_second = attr.ib()
    camera = attr.ib()
    events = attr.ib()
    trap = attr.ib()
    tracks = attr.ib()
    ffc_frames = attr.ib()
    country_code = attr.ib()
    frame_temp_median = attr.ib(default=None)

    def get_samples(self):
        samples = []
        for t in self.tracks:
            samples.extend(t.samples)
        return samples


class TrackHeader:
    """Header for track."""

    def __init__(
        self,
        clip_id,
        track_id,
        label,
        num_frames,
        regions,
        start_frame,
        ffc_frames=None,
        sample_frames_indices=None,
        station_id=None,
        start_time=None,
        source_file=None,
        camera=None,
        confidence=None,
        human_tags=None,
        remapped_lbl=None,
        mega_missed_regions=None,
        skip_ffc=True,
        fp_frames=None,
    ):

        self.fp_frames = fp_frames
        self.start_time = start_time
        # regions that megadetector found nothing in
        self.mega_missed_regions = mega_missed_regions
        self.station_id = station_id
        self.clip_id = clip_id
        self.source_file = source_file
        self.human_tags = human_tags
        self.confidence = confidence
        self.filtered_stats = {"segment_mass": 0}
        # reference to clip this segment came from
        # reference to track this segment came from
        self.track_id = track_id
        # label for this track
        # date and time of the start of the track
        self.start_frame = np.uint16(start_frame)

        self.original_label = label
        self.remapped_label = remapped_lbl

        # thermal reference point for each frame.
        # self.frame_temp_median = np.uint16(frame_temp_median)
        # tracking frame movements for each frame, array of tuples (x-vel, y-vel)
        self.frame_velocity = None
        # original tracking bounds
        self.regions_by_frame = regions
        # what fraction of pixels are from out of bounds
        self.frame_crop = None
        self.num_frames = num_frames
        self.important_predicted = 0
        mass_history = [
            region.mass
            for region in self.regions_by_frame.values()
            if region.mass > 0
            and (
                ffc_frames is None
                or skip_ffc is False
                or region.frame_number not in ffc_frames
            )
            and not region.blank
            and region.width > 0
            and region.height > 0
        ]
        self.has_no_mass = np.sum(mass_history) == 0
        if len(mass_history) == 0:
            self.lower_mass = 0
            self.upper_mass = 0
            self.median_mass = 0
            self.mean_mass = 0
        else:
            self.lower_mass = np.uint16(np.percentile(mass_history, q=25))
            self.upper_mass = np.uint16(np.percentile(mass_history, q=75))
            self.median_mass = np.uint16(np.median(mass_history))
            self.mean_mass = np.uint16(np.mean(mass_history))
        self.samples = []

    @property
    def label(self):
        return (
            self.original_label if self.remapped_label is None else self.remapped_label
        )

    @property
    def bounds_history(self):
        regions = list(self.regions_by_frame.values())
        regions = sorted(regions, key=lambda r: r.frame_number)
        return regions

    def toJSON(self, clip_meta):
        meta_dict = {}
        ffc_frames = clip_meta.get("ffc_frames", [])
        meta_dict["ffc_frames"] = ffc_frames
        meta_dict["clip_id"] = int(self.clip_id)
        meta_dict["track_id"] = int(self.track_id)
        meta_dict["camera"] = self.camera
        meta_dict["num_frames"] = int(self.num_frames)
        meta_dict["frames_per_second"] = int(self.frames_per_second)
        meta_dict["track_bounds"] = self.regions
        meta_dict["start_frame"] = int(self.start_frame)
        if self.location is not None:
            try:
                meta_dict["location_hash"] = "{}{}".format(
                    hash(self.location[0]), hash(self.location[1])
                )
            except:
                pass
        meta_dict["label"] = self.label

        return json.dumps(meta_dict, indent=3, cls=tools.CustomJSONEncoder)

    def add_sample(self, sample):
        self.samples.append(sample)

    def calculate_sample_frames(
        self,
        min_mass=None,
        max_mass=None,
        ffc_frames=None,
        skip_last=None,
        max_frames=None,
    ):
        crop_rectangle = Rectangle(2, 2, 160 - 2 * 2, 140 - 2 * 2)

        logging.debug(
            "Calculating sample with min %s and max %s ffc %s and skip %s",
            min_mass,
            max_mass,
            ffc_frames,
            skip_last,
        )
        frame_numbers = list(self.regions_by_frame.keys())
        previous_mass = None

        if skip_last is not None:
            skip_x = int(len(frame_numbers) * skip_last)
            frame_numbers = frame_numbers[:-skip_x]

        frame_numbers.sort()
        for frame_num in frame_numbers:
            region = self.regions_by_frame[frame_num]

            if (
                region.mass == 0
                or region.blank
                or region.width <= 0
                or region.height <= 0
            ):
                continue
            if ffc_frames is not None and frame_num in ffc_frames:
                continue

            if (
                self.mega_missed_regions is not None
                and frame_num in self.mega_missed_regions
            ):
                continue

            if min_mass is not None and region.mass < min_mass:
                continue
            if max_mass is not None and region.mass > max_mass:
                continue
            # dont use regions on the edge if the mass deviates too much from the last known good mass
            region.set_is_along_border(crop_rectangle)
            if region.is_along_border:
                if previous_mass is not None:
                    previous_mass_thresh = previous_mass * 0.1
                    if (abs(previous_mass - region.mass)) >= previous_mass_thresh:
                        continue
            else:
                previous_mass = region.mass
            f = FrameSample(
                self.clip_id,
                self.track_id,
                region.frame_number,
                self.label,
                None,
                region,
                source_file=self.source_file,
                weight=1,
                station_id=self.station_id,
                track_median_mass=self.median_mass,
            )
            self.samples.append(f)
        if max_frames is not None and len(self.samples) > max_frames:
            self.samples = np.random.choice(self.samples, max_frames, replace=False)

    def remove_sample(self, f):
        self.samples.remove(f)

    #
    # def get_sample_frame(self, i=0, remove=False):
    #     if len(self.sample_frames) == 0:
    #         return None
    #     f = self.sample_frames[i]
    #
    #     if remove:
    #         del self.sample_frames[i]
    #     return f

    @property
    def frame_weight(self):
        return 1 / self.num_sample_frames

    #
    # @property
    # def num_samples(self):
    #     return len(self.sample_frames)

    def calculate_frame_crop(self):
        # frames are always square, but bounding rect may not be, so to see how much we clipped I need to create a square
        # bounded rect and check it against frame size.
        self.frame_crop = []
        for rect in self.regions:
            rx, ry = rect.mid_x, rect.mid_y
            size = max(rect.width, rect.height)
            adjusted_rect = tools.Rectangle(rx - size / 2, ry - size / 2, size, size)
            self.frame_crop.append(
                get_cropped_fraction(adjusted_rect, self.res_x, self.res_y)
            )

    def calculate_velocity(self):
        frame_center = [region.mid for region in self.track_bounds]
        self.frame_velocity = []
        prev = None
        for x, y in frame_center:
            if prev is None:
                self.frame_velocity.append((0.0, 0.0))
            else:
                self.frame_velocity.append((x - prev[0], y - prev[1]))
            prev = (x, y)

    def get_segments(
        self,
        segment_width,
        segment_frame_spacing=9,
        segment_types=[SegmentType.ALL_RANDOM],
        segment_min_mass=None,
        repeats=1,
        max_segments=None,
        dont_filter=False,
        skip_ffc=True,
        ffc_frames=None,
        location=None,
        segment_frames=None,
        from_last=None,
        frame_min_mass=None,
        filter_by_fp=True,
        min_segments=None,
    ):
        if segment_frames is not None:
            raise Exception("Have not implement this path")
        min_frames = segment_width / 4.0
        if self.label in BuildConfig.NO_MIN_FRAMES:
            # try and always get one for these
            min_frames = 0
            if min_segments is None:
                min_segments = 1

        # in python3.7+ can just take the values and it guarantees order it was added to dict
        regions = self.bounds_history
        self.samples, filtered_stats = get_segments(
            self.clip_id,
            self.track_id,
            self.start_frame,
            segment_frame_spacing=segment_frame_spacing,
            segment_width=segment_width,
            label=self.label,
            regions=np.array(regions),
            segment_min_mass=segment_min_mass,
            ffc_frames=ffc_frames,
            lower_mass=self.lower_mass,
            repeats=repeats,
            min_frames=min_frames,
            segment_types=segment_types,
            max_segments=max_segments,
            station_id=self.station_id,
            source_file=self.source_file,
            dont_filter=dont_filter,
            skip_ffc=skip_ffc,
            frame_min_mass=frame_min_mass,
            fp_frames=self.fp_frames if filter_by_fp else None,
            rec_time=self.start_time,
            min_segments=min_segments,
        )
        self.filtered_stats.update(filtered_stats)
        # GP could get this from the tracks when writing
        # but might be best to keep samples independent for ease
        for s in self.samples:
            s._track_median_mass = self.median_mass
        return self.samples

    @property
    def camera_id(self):
        """Unique name of this track."""
        return "{}-{}".format(self.camera, self.location)

    @property
    def bin_id(self):
        """Unique name of this track."""
        return "{}-{}".format(self.clip_id, self.location_id)

    @property
    def weight(self):
        """Returns total weight for all segments in this track"""
        return sum(segment.weight for segment in self.segments)

    @property
    def unique_id(self):
        return "{}-{}".format(self.clip_id, self.track_id)

    @staticmethod
    def from_meta(source_file, clip_id, clip_meta, track_meta, predictions=None):
        """Creates a track header from given metadata."""
        correct_prediction = track_meta.get("correct_prediction", None)
        rec_time = dateutil.parser.parse(clip_meta["start_time"])
        # end_time = dateutil.parser.parse(track_meta["end_time"])
        location = clip_meta.get("location")
        camera = clip_meta.get("device_id")
        # clip_meta["device"]
        station_id = clip_meta.get("station_id")
        frames_per_second = clip_meta.get("frames_per_second", FRAMES_PER_SECOND)
        # get the reference levels from clip_meta and load them into the track.
        track_start_frame = int(round(track_meta["start_frame"]))
        track_end_frame = int(round(track_meta["end_frame"]))
        num_frames = track_end_frame - track_start_frame + 1
        frame_temp_median = np.float32(
            clip_meta["frame_temp_median"][
                track_start_frame : num_frames + track_start_frame
            ]
        )

        ffc_frames = clip_meta.get("ffc_frames", [])
        sample_frames = None
        # sample_frames = track_meta.get("sample_frames")
        # if sample_frames is not None:
        #     sample_frames = sample_frames + track_start_frame
        regions = {}
        f_i = 0
        first_region = None
        for bounds in track_meta["regions"]:
            r = Region.region_from_array(bounds)
            if first_region is None:
                first_region = r
            regions[r.frame_number] = r
            f_i += 1
        if track_start_frame != first_region.frame_number:
            logging.warn(
                "Have wrong region or start frame for %s clip %s track %s adjusting regions first region %s becomes %s",
                source_file,
                clip_id,
                track_meta["id"],
                first_region.frame_number,
                track_start_frame,
            )

            region_list = list(regions.values())
            regions = {}
            frame_number = track_start_frame
            for r in region_list:
                r = Region.region_from_array(bounds)
                r.frame_number = frame_number
                frame_number += 1
                regions[r.frame_number] = r
        header = TrackHeader(
            clip_id=int(clip_id),
            track_id=int(track_meta["id"]),
            label=track_meta["human_tag"],
            num_frames=num_frames,
            location=location,
            regions=regions,
            frame_temp_median=frame_temp_median,
            frames_per_second=frames_per_second,
            predictions=predictions,
            correct_prediction=correct_prediction,
            start_frame=track_start_frame,
            res_x=clip_meta.get("res_x", CPTV_FILE_WIDTH),
            res_y=clip_meta.get("res_y", CPTV_FILE_HEIGHT),
            ffc_frames=ffc_frames,
            sample_frames_indices=sample_frames,
            station_id=station_id,
            rec_time=rec_time,
            source_file=source_file,
            camera=camera,
        )
        return header

    def __repr__(self):
        return self.unique_id

    def get_id(self):
        return self.track_id


class Camera:
    def __init__(self, camera):
        self.label_to_bins = {}
        self.label_to_tracks = {}
        self.label_to_samples = {}

        self.bins = {}
        self.camera = camera
        self.segment_sum = 0
        self.segments = 0
        self.label_frames = {}
        # used to sample bins
        self.bin_i = -1

    def label_tracks(self, label):
        return len(self.label_to_bins[label])

    def sample_frame(self, label):
        bins = self.label_to_bins[label]
        if len(bins) == 0:
            return None, None
        self.bin_i += 1
        self.bin_i = self.bin_i % len(bins)

        bin_id = bins[self.bin_i]
        track = self.bins[bin_id][0]
        f = track.get_sample_frame()
        if len(track.sample_frames) == 0 or f is None:
            del bins[self.bin_i]
            del self.bins[bin_id]

        return track, f

    def label_track_count(
        self,
        label,
        max_segments_per_track=None,
    ):
        if label not in self.label_to_tracks:
            return 0
        tracks = self.label_to_tracks[label].values()
        return len(tracks)

    def label_segment_count(
        self,
        label,
        max_segments_per_track=None,
    ):
        if label not in self.label_to_tracks:
            return 0
        tracks = self.label_to_tracks[label].values()
        frames = 0
        for track in tracks:
            if max_segments_per_track:
                frames += max(len(track.segments), max_segments_per_track)
            else:
                frames += len(track.segments)

        return frames

    def label_frame_count(
        self,
        label,
        max_frames_per_track=None,
    ):
        if label not in self.label_to_tracks:
            return 0
        tracks = self.label_to_tracks[label].values()
        frames = 0
        for track in tracks:
            if max_frames_per_track:
                frames += max(len(track.sample_frames), max_frames_per_track)
            else:
                frames += len(track.sample_frames)

        return frames

    def remove_label(self, label):
        if label in self.label_to_tracks:
            tracks = list(self.label_to_tracks[label].values())
            while len(tracks) > 0:
                track = tracks.pop()
                self.remove_track(track)
            del self.label_to_tracks[label]
            del self.label_to_bins[label]
            del self.label_frames[label]

    def remove_track(self, track):
        self.segments -= 1
        self.segment_sum -= len(track.segments)
        del self.label_to_tracks[track.label][track.unique_id]
        if track.bin_id in self.label_to_bins[track.label]:
            self.label_to_bins[track.label].remove(track.bin_id)
        self.label_frames[track.label] -= len(track.sample_frames)

    def add_track(self, track_header):
        tracks = self.label_to_tracks.setdefault(track_header.label, {})
        tracks[track_header.unique_id] = track_header
        if track_header.bin_id not in self.bins:
            self.bins[track_header.bin_id] = []

        if track_header.label not in self.label_to_bins:
            self.label_to_bins[track_header.label] = []
            self.label_frames[track_header.label] = 0

        if track_header.bin_id not in self.label_to_bins[track_header.label]:
            self.label_to_bins[track_header.label].append(track_header.bin_id)

        self.bins[track_header.bin_id].append(track_header)
        self.label_frames[track_header.label] += len(track_header.sample_frames)

        segment_length = len(track_header.segments)
        self.segment_sum += segment_length
        self.segments += 1

    def remove_sample(self, sample):
        del self.label_to_samples[sample.label][sample.unique_id]
        if track.bin_id in self.label_to_bins[sample.label]:
            self.label_to_bins[sample.label].remove(sample.bin_id)
        self.label_frames[sample.label] -= len(sample.sample_frames)

    def add_sample(self, sample):
        samples = self.label_to_samples.setdefault(sample.label, {})
        samples[sample.unique_id] = sample
        if sample.bin_id not in self.bins:
            self.bins[sample.bin_id] = []

        if sample.label not in self.label_to_bins:
            self.label_to_bins[sample.label] = []
            self.label_frames[sample.label] = 0

        if sample.bin_id not in self.label_to_bins[sample.label]:
            self.label_to_bins[sample.label].append(sample.bin_id)

        self.bins[sample.bin_id].append(sample)


class FrameSample(Sample):
    _frame_id = 1

    def __init__(
        self,
        clip_id,
        track_id,
        frame_num,
        label,
        velocity,
        region,
        weight,
        source_file,
        augment=False,
        track_median_mass=None,
        station_id=None,
    ):
        super().__init__(label)
        self.clip_id = clip_id
        self.id = FrameSample._frame_id
        FrameSample._frame_id += 1

        self.track_id = track_id
        self.frame_number = frame_num
        self.velocity = velocity
        self.region = region
        self.weight = weight
        self.augment = augment
        self._source_file = source_file
        self._track_median_mass = track_median_mass
        self.station_id = station_id

    @property
    def filtered(self):
        return False

    @property
    def track_median_mass(self):
        return self._track_median_mass

    @property
    def source_file(self):
        return self._source_file

    def copy(self):
        f = FrameSample(
            clip_id=self.clip_id,
            track_id=self.track_id,
            frame_num=self.frame_number,
            label=self.label,
            temp_median=self.temp_median,
            velocity=self.velocity,
            region=self.region,
            weight=self.weight,
            camera=self.camera,
            start_time=self.start_time,
            augment=self.augment,
            source_file=self.source_file,
            station_id=self.station_id,
            rec_time=self.rec_time,
            track_median_mass=self.track_median_mass,
        )
        return f

    @property
    def frame_numbers(self):
        return [self.frame_number]

    @property
    def mass(self):
        return self.region.mass

    @property
    def sample_weight(self):
        return self.weight

    @property
    def unique_track_id(self):
        return "{}-{}".format(self.clip_id, self.track_id)

    @property
    def track_bounds(self):
        return [self.region]

    @property
    def frame_indices(self):
        return [self.frame_number]

    @property
    def unique_id(self):
        return f"{self.clip_id}-{self.track_id}-{self.frame_number}"

    @property
    def bin_id(self):
        """Unique name of this segments track."""
        # break into 50 frame keys, since we dont have much data this means multiple sets can have same clip
        i = int(self.frame_number / 50)
        return f"{self.clip_id}"


class SegmentHeader(Sample):
    """Header for segment."""

    _segment_id = 1

    def __init__(
        self,
        clip_id,
        track_id,
        start_frame,
        frames,
        weight,
        mass,
        label,
        regions,
        frame_indices=None,
        movement_data=None,
        best_mass=False,
        top_mass=False,
        start_time=None,
        camera=None,
        location=None,
        station_id=None,
        rec_time=None,
        source_file=None,
        filtered=False,
        track_median_mass=None,
    ):
        super().__init__(label)
        self.filtered = filtered
        self.rec_time = rec_time
        self.location = location
        self.station_id = station_id
        self.movement_data = movement_data
        self.top_mass = top_mass
        self.best_mass = best_mass
        self.id = SegmentHeader._segment_id
        SegmentHeader._segment_id += 1
        # reference to track this segment came from
        self.clip_id = clip_id
        self.track_id = track_id
        self.frame_numbers = np.uint16(frame_indices)
        self.start_time = start_time
        self.regions = regions

        self.start_frame = start_frame
        # length of segment in frames
        self.frames = np.uint16(frames)
        # relative weight of the segment (higher is sampled more often)
        self.weight = np.float16(weight)

        self._mass = np.uint16(mass)
        self.camera = camera
        self._source_file = source_file
        self._track_median_mass = track_median_mass

    @property
    def track_median_mass(self):
        return self._track_median_mass

    @property
    def source_file(self):
        return self._source_file

    @property
    def mass(self):
        return self._mass

    @property
    def sample_weight(self):
        return self.weight

    @property
    def track_bounds(self):
        return self.regions

    @property
    def frame_indices(self):
        return self.frame_numbers

    @property
    def avg_mass(self):
        return self.mass / self.frames

    @property
    def unique_track_id(self):
        # reference to clip this segment came from
        return "{}-{}".format(self.clip_id, self.track_id)

    @property
    def name(self):
        """Unique name of this segment."""
        return self.clip_id + "-" + str(self.track_id) + "-" + str(self.start_frame)

    @property
    def frame_velocity(self):
        # tracking frame velocity for each frame.
        return self.track.frame_velocity[
            self.start_frame : self.start_frame + self.frames
        ]

    @property
    def frame_crop(self):
        # how much each frame has been cropped.
        return self.track.frame_crop[self.start_frame : self.start_frame + self.frames]

    @property
    def end_frame(self):
        """end frame of segment"""
        return self.start_frame + self.frames

    @property
    def bin_id(self):
        """Unique name of this segments track."""
        return f"{self.station_id}"

    def __str__(self):
        return "{0} label {1} offset:{2} weight:{3:.1f}".format(
            self.unique_track_id, self.label, self.start_frame, self.weight
        )

    @property
    def unique_id(self):
        return self.id

    def get_data(self, db):
        crop_rectangle = tools.Rectangle(2, 2, 160 - 2 * 2, 140 - 2 * 2)

        try:
            background = db.get_clip_background(self.clip_id)
            frames = db.get_track(
                self.clip_id,
                self.track_id,
                original=False,
                frame_numbers=self.frame_numbers - self.start_frame,
            )

            thermals = []  # np.empty(len(frames), dtype=object)
            filtered = []  # np.empty(len(frames), dtype=object)

            for i, frame in enumerate(frames):
                frame.float_arrays()
                frame.filtered = frame.thermal - frame.region.subimage(background)
                temp_index = np.where(self.frame_numbers == frame.frame_number)[0][0]
                temp = self.frame_temp_median[temp_index]
                frame.resize_with_aspect((32, 32), crop_rectangle, keep_edge=True)
                frame.thermal -= temp
                np.clip(frame.thermal, a_min=0, a_max=None, out=frame.thermal)

                frame.thermal, stats = imageprocessing.normalize(
                    frame.thermal, new_max=255
                )
                if not stats[0]:
                    frame.thermal = np.zeros((frame.thermal.shape))
                    # continue
                frame.filtered, stats = imageprocessing.normalize(
                    frame.filtered, new_max=255
                )
                if not stats[0]:
                    frame.filtered = np.zeros((frame.filtered.shape))

                # continue
                # frame.filtered =
                filtered.append(frame.filtered)
                thermals.append(frame.thermal)
            thermals = np.array(thermals)
            filtered = np.array(filtered)
            # thermal, success = imageprocessing.square_clip(thermals, 5, (32, 32))
            # if not success:
            #     logging.warn("Error making thermal square clip %s", self.clip_id)
            #     return None
            # filtered, success = imageprocessing.square_clip(filtered, 5, (32, 32))
            # if not success:
            #     logging.warn("Error making filtered square clip %s", filtered)
            #
            #     return None
        except:
            logging.error("Cant get segment %s", self, exc_info=True)
            raise "EX"
            return None
        return thermals, filtered


def get_cropped_fraction(region: Rectangle, width, height):
    """Returns the fraction regions mass outside the rect ((0,0), (width, height)"""
    bounds = tools.Rectangle(0, 0, width - 1, height - 1)
    return 1 - (bounds.overlap_area(region) / region.area)


def get_movement_data(regions):
    areas = np.array([region.area for region in regions])
    centrex = np.array([region.mid_x for region in regions])
    centrey = np.array([region.mid_y for region in regions])
    xv = np.hstack((0, centrex[1:] - centrex[:-1]))
    yv = np.hstack((0, centrey[1:] - centrey[:-1]))
    axv = xv / areas**0.5
    ayv = yv / areas**0.5
    bounds = np.array([region.to_ltrb() for region in regions])
    mass = np.array([region.mass for region in regions])

    return np.hstack((bounds, np.vstack((mass, xv, yv, axv, ayv)).T))


# should use trackheader / track in creationg interface out common properties
def get_segments(
    clip_id,
    track_id,
    start_frame,
    regions,
    segment_width=25,
    segment_frame_spacing=9,
    label=None,
    segment_min_mass=None,
    ffc_frames=[],
    lower_mass=0,
    repeats=1,
    min_frames=None,
    segment_types=[SegmentType.ALL_RANDOM],
    max_segments=None,
    location=None,
    station_id=None,
    camera=None,
    rec_time=None,
    source_file=None,
    dont_filter=False,
    skip_ffc=True,
    frame_min_mass=None,
    fp_frames=None,
    repeat_frame_indices=True,
    min_segments=None,
):
    if min_frames is None:
        min_frames = segment_width / 4.0
    segments = []
    mass_history = np.uint16([region.mass for region in regions])
    filtered_stats = {"segment_mass": 0, "too short": 0}
    has_no_mass = np.sum(mass_history) == 0

    for segment_type in segment_types:
        s_min_mass = segment_min_mass
        if segment_type == SegmentType.ALL_RANDOM_NOMIN:
            s_min_mass = None

        frame_indices = [
            region.frame_number
            for region in regions
            if (has_no_mass or region.mass > 0)
            and (
                ffc_frames is None
                or skip_ffc is False
                or region.frame_number not in ffc_frames
            )
            and not region.blank
            and region.width > 0
            and region.height > 0
            and (
                (has_no_mass or frame_min_mass is None) or region.mass >= frame_min_mass
            )
        ]
        if fp_frames is not None and label not in FP_LABELS:
            frame_indices = [f for f in frame_indices if f not in fp_frames]
        if len(frame_indices) == 0:
            logging.warn("Nothing to load for %s - %s", clip_id, track_id)
            return [], filtered_stats
        if s_min_mass is not None:
            s_min_mass = min(
                s_min_mass,
                np.median(mass_history[frame_indices - start_frame]),
            )
        else:
            s_min_mass = 1
            # remove blank frames

        if segment_type == SegmentType.TOP_RANDOM:
            # take top 50 mass frames
            frame_indices = sorted(
                frame_indices,
                key=lambda f_i: mass_history[f_i - start_frame],
                reverse=True,
            )
            frame_indices = frame_indices[:50]
            frame_indices.sort()
        if segment_type == SegmentType.TOP_SEQUENTIAL:
            new_segments, filtered = get_top_mass_segments(
                clip_id,
                track_id,
                label,
                camera,
                segment_width,
                segment_frame_spacing,
                mass_history,
                ffc_frames,
                regions,
                start_frame,
                lower_mass,
                s_min_mass,
                source_file=source_file,
            )
            segments.extend(new_segments)
            filtered_stats.merge(filtered)
            continue
        if len(frame_indices) < min_frames and (
            min_segments == 0 or min_segments is None
        ):
            filtered_stats["too short"] += 1
            continue

        frame_indices = np.array(frame_indices)
        segment_count = max(1, len(frame_indices) // segment_frame_spacing)
        segment_count = int(segment_count)
        mask_length = 25
        # probably only counts for all random
        if max_segments is not None and segment_type not in [SegmentType.ALL_SECTIONS]:
            segment_count = min(max_segments, segment_count)
            # adjust size of mask if we take less segments
            mask_length = max(mask_length, len(frame_indices) // segment_count)
        # take any segment_width frames, this could be done each epoch
        whole_indices = frame_indices
        random_frames = segment_type in [
            SegmentType.IMPORTANT_RANDOM,
            SegmentType.ALL_RANDOM,
            SegmentType.ALL_RANDOM_NOMIN,
            SegmentType.TOP_RANDOM,
            SegmentType.ALL_RANDOM_MASKED,
            None,
        ]
        for _ in range(repeats):
            if segment_type == SegmentType.ALL_RANDOM_MASKED:
                segment_indices = np.arange(len(regions))
                all_frames = np.arange(len(regions)) + start_frame

                available_indices = np.full(len(regions), False)
                available_indices[frame_indices - start_frame] = True
                # use all frames and make out onces filtered by other criteria like mass and blank etc
                # this way we are always making as 25 frame period
            if segment_type != SegmentType.ALL_RANDOM_MASKED or len(whole_indices) < 40:
                frame_indices = whole_indices.copy()

                if random_frames:
                    # random_frames and not random_sections:
                    np.random.shuffle(frame_indices)
            for i in range(segment_count):
                if segment_type == SegmentType.ALL_RANDOM_MASKED:
                    if len(whole_indices) < 40:
                        frame_indices = segment_indices[available_indices]
                    else:
                        mask = available_indices.copy()
                        mask_start = i * mask_length
                        mask[mask_start : mask_start + mask_length] = False

                        frame_indices = segment_indices[mask]
                        frame_indices = np.uint32(frame_indices)
                        np.random.shuffle(frame_indices)

                # always get atleast one segment, not doing annymore
                if (
                    len(frame_indices) == 0
                    or min_segments is None
                    or len(segments) >= min_segments
                ):
                    if (
                        len(frame_indices) < segment_width / 2.0 and len(segments) > 1
                    ) or len(frame_indices) < segment_width / 4:
                        break

                if segment_type == SegmentType.ALL_SECTIONS:
                    # random frames from section 2.2 * segment_width
                    section = frame_indices[: int(segment_width * 2.2)]

                    indices = np.random.choice(
                        len(section),
                        min(segment_width, len(section)),
                        replace=False,
                    )
                    frames = section[indices]
                    # might need to change that gp 11/05 - 2024
                    frame_indices = frame_indices[segment_width:]
                elif segment_type == SegmentType.ALL_RANDOM_MASKED:
                    indices = frame_indices[:segment_width]
                    available_indices[indices] = False
                    frames = all_frames[indices]
                elif random_frames:
                    # frame indices already randomized so just need to grab some
                    frames = frame_indices[:segment_width]
                    frame_indices = frame_indices[segment_width:]
                else:
                    segment_start = i * segment_frame_spacing
                    segment_end = segment_start + segment_width
                    segment_end = min(len(frame_indices), segment_end)
                    frames = frame_indices[segment_start:segment_end]

                remaining = segment_width - len(frames)
                # sample another same frames again if need be
                if remaining > 0:
                    extra_frames = np.random.choice(
                        frames,
                        min(remaining, len(frames)),
                        replace=False,
                    )
                    frames = np.concatenate([frames, extra_frames])
                frames.sort()
                relative_frames = frames - start_frame
                mass_slice = mass_history[relative_frames]
                segment_mass = np.sum(mass_slice)
                segment_avg_mass = segment_mass / len(mass_slice)
                filtered = False
                if s_min_mass and segment_avg_mass < s_min_mass:
                    if dont_filter:
                        filtered = True
                    else:
                        filtered_stats["segment_mass"] += 1
                        continue

                # temp_slice = frame_temp_median[relative_frames]
                region_slice = regions[relative_frames]
                movement_data = None
                if segment_avg_mass < 50:
                    segment_weight_factor = 0.75
                elif segment_avg_mass < 100:
                    segment_weight_factor = 1
                else:
                    segment_weight_factor = 1.2

                for z, f in enumerate(frames):
                    assert region_slice[z].frame_number == f

                if repeat_frame_indices:
                    # i think this can be default, means we dont need to handle
                    # short segments elsewhere
                    if len(frames) < segment_width:
                        extra_samples = np.random.choice(
                            frames, segment_width - len(frames)
                        )
                        frames = list(frames)
                        frames.extend(extra_samples)
                        frames.sort()

                segment = SegmentHeader(
                    clip_id,
                    track_id,
                    start_frame=start_frame,
                    frames=segment_width,
                    weight=segment_weight_factor,
                    mass=segment_mass,
                    label=label,
                    regions=region_slice,
                    frame_indices=frames,
                    movement_data=movement_data,
                    camera=camera,
                    location=location,
                    station_id=station_id,
                    rec_time=rec_time,
                    source_file=source_file,
                    filtered=filtered,
                )
                segments.append(segment)
    return segments, filtered_stats


def get_top_mass_segments(
    clip_id,
    track_id,
    label,
    camera,
    segment_width,
    segment_frame_spacing,
    mass_history,
    ffc_frames,
    regions,
    start_frame,
    lower_mass,
    segment_min_mass,
    source_file=None,
):
    # Needs re testing if want to use
    logging.warning("Top mass segments hasn't been tested may not work")
    filtered_stats = {"segment_mass": 0, "too short": 0}

    segments = []
    segment_count = max(1, len(regions) // segment_frame_spacing)
    segment_count = int(segment_count)

    segment_mass = []
    for i in range(max(1, len(mass_history) - segment_width)):
        contains_ffc = False
        for z in range(segment_width):
            if (z + i + start_frame) in ffc_frames:
                contains_ffc = True
                break
        if contains_ffc:
            continue
        mass = np.sum(mass_history[i : i + segment_width])
        segment_mass.append((i, mass))

    sorted_mass = sorted(segment_mass, key=lambda x: x[1], reverse=True)
    best_mass = True
    segment_count = max(1, len(regions) // segment_frame_spacing)
    segment_count = int(min(len(sorted_mass), segment_count))

    for _ in range(segment_count):
        segment_info = sorted_mass[0]
        index = segment_info[0]
        avg_mass = segment_info[1] / segment_width
        if not best_mass and (avg_mass < lower_mass or avg_mass < segment_min_mass):
            break
        movement_data = get_movement_data(regions[index : index + segment_width])
        width = min(segment_width, len(regions))
        frames = np.arange(width) + index
        segment = SegmentHeader(
            clip_id,
            track_id,
            start_frame=start_frame,
            frames=segment_width,
            weight=1,
            mass=segment_info[1],
            label=label,
            regions=regions[frames],
            frame_indices=frames + start_frame,
            movement_data=movement_data,
            best_mass=best_mass,
            top_mass=True,
            camera=camera,
            source_file=source_file,
        )
        best_mass = False
        segments.append(segment)
        # remove those that start within this segment
        sorted_mass = [
            mass_info
            for mass_info in sorted_mass
            if mass_info[0] <= (index - segment_width / 3 * 2)
            or mass_info[0] >= (index + segment_width / 3 * 2)
        ]
        if len(segments) == segment_count or len(sorted_mass) == 0:
            break
    return segments, filtered_stats


# GP Not using at the moment, but handy for more complex training


class TrackingSample(Sample):
    _s_id = 1

    def __init__(
        self,
        clip_id,
        track_id,
        frame_num,
        labels,
        temp_median,
        region,
        camera,
        filename,
    ):
        self.id = TrackingSample._s_id
        TrackingSample._s_id += 1
        self.clip_id = clip_id
        self.filename = filename
        self.track_id = track_id
        # needed for original frame will change this GP
        self.frame_number = frame_num
        self.camera = camera
        if isinstance(labels, str):
            self.labels = [labels]
        else:
            self.labels = labels
        self.temp_median = temp_median
        if region is None:
            self.regions = []
        else:
            self.regions = [region]

    # asusming only one label for now GP should change
    @property
    def label(self):
        return self.labels[0]

    @property
    def region(self):
        return self.regions[0]

    def add_sample(self, f):
        self.labels.append(f.label)
        self.regions.append(f.region)

    def filename(self):
        return f"{self.clip_id}-{self.track_id}-{self.frame_number}"

    @classmethod
    def from_sample(cls, f):
        return cls(
            f.clip_id, f.track_id, f.frame_number, [f.label], f.temp_median, [f.region]
        )

    @property
    def unique_track_id(self):
        return "{}-{}".format(self.clip_id, self.track_id)

    @property
    def track_bounds(self):
        return self.regions

    @property
    def frame_indices(self):
        return [self.frame_number]

    @property
    def sample_weight(self):
        return 1

    @property
    def unique_id(self):
        return f"{self.clip_id}-{self.track_id}-{self.frame_number}"

    @property
    def bin_id(self):
        """Unique name of this segments track."""
        frames_numbers = int(self.frame_number / 20)
        # try and segment framesso 20 frames are in the same bin
        return f"{self.clip_id}-{self.track_id}-{frames_numbers}"
        # this should be used but dont have much data
        # return f"{self.clip_id}-{self.track_id}"



import tensorflow as tf


# https://arxiv.org/pdf/1605.07146.pdf
def WRResNet(X_input, depth=22, k=4, groups=2):
    filters = [16, 16 * k, 32 * k, 64 * k]
    # Define the input as a tensor with shape input_shape
    # X_input = tf.keras.Input(input_shape, name="wr-input")
    n = int((depth - 4) / 6)
    for stage, f in enumerate(filters):
        if stage == 0:
            X = tf.keras.layers.Conv2D(
                f,
                (3, 3),
                strides=1,
                padding="same",
                name=f"conv1_{stage+1}",
                kernel_initializer=tf.keras.initializers.GlorotUniform(seed=0),
                groups=groups,
            )(X_input)
        else:
            X = wr_block(
                X,
                3,
                (f, f),
                stage=stage + 1,
                block="b",
                stride=stage,
                depth=n,
                groups=groups,
            )
    #
    X = tf.keras.layers.BatchNormalization(axis=3, name="final_bn")(X)
    X = tf.keras.layers.Activation("relu")(X)
    model = tf.keras.Model(inputs=X_input, outputs=X, name="WRResNet")
    return model


def wr_block(X, f, filters, stage, block, stride=1, depth=1, groups=2):
    s_block = f"{block}0"

    X = basic_block(X, f, filters, stage, s_block, stride, groups=groups)
    for d in range(depth - 1):
        s_block = f"{block}{d+1}"
        X = basic_block(X, f, filters, stage, s_block, 1, groups=groups)
    return X


def basic_block(X, f, filters, stage, block, stride=1, groups=2):
    # defining name basis
    conv_name_base = "res" + str(stage) + block + "_branch"
    bn_name_base = "bn" + str(stage) + block + "_branch"

    # Retrieve Filters
    F1, F2 = filters
    # Save the input value. You'll need this later to add back to the main path.
    X_shortcut = X
    # First component of main path
    X = tf.keras.layers.BatchNormalization(axis=3, name=bn_name_base + "2a")(X)
    X = tf.keras.layers.Activation("relu")(X)
    X = tf.keras.layers.Conv2D(
        filters=F1,
        kernel_size=(3, 3),
        strides=(stride, stride),
        padding="same",
        name=conv_name_base + "2a",
        kernel_initializer=tf.keras.initializers.GlorotUniform(seed=0),
        groups=groups,
    )(X)

    X = tf.keras.layers.Dropout(rate=0.1)(X)
    # , training=istraining_ph)

    # Second component of main path
    X = tf.keras.layers.BatchNormalization(axis=3, name=bn_name_base + "2b")(X)
    X = tf.keras.layers.Activation("relu")(X)
    X = tf.keras.layers.Conv2D(
        filters=F2,
        kernel_size=(3, 3),
        strides=1,
        padding="same",
        name=conv_name_base + "2b",
        kernel_initializer=tf.keras.initializers.GlorotUniform(seed=0),
        groups=groups,
    )(X)
    if X.shape[-1] == X_shortcut.shape[-1]:
        X_shortcut = tf.keras.layers.Identity()(X_shortcut)
    else:
        X_shortcut = tf.keras.layers.Conv2D(
            X.shape[-1],
            strides=(stride, stride),
            kernel_size=1,
            groups=groups,
        )(X_shortcut)
    # Final step: Add shortcut value to main path, and pass it through a RELU activation
    X = tf.keras.layers.Add()([X, X_shortcut])
    X = tf.keras.layers.Activation("relu")(X)
    return X


def main():
    wr = WRResNet(tf.keras.Input(shape=(160, 160, 2), name="input"), depth=22, k=4)
    x = tf.keras.layers.GlobalAveragePooling2D()(wr.output)

    x = tf.keras.layers.Dropout(0.3)(x)
    output = tf.keras.layers.Dense(17, activation="sigmoid", name="prediction")(x)
    model = tf.keras.models.Model(wr.inputs, outputs=output)

    model.summary()


if __name__ == "__main__":
    main()



import attr
import cv2
import numpy as np
from ml_tools.tools import get_clipped_flow
from ml_tools.imageprocessing import resize_cv, rotate, normalize, resize_and_pad
import enum


class TrackChannels(enum.Enum):
    """Indexes to channels in track."""

    thermal = 0
    filtered = 1
    flow_h = 2
    flow_v = 3
    mask = 4
    flow = 5


@attr.s(slots=True, eq=False)
class Frame:
    thermal = attr.ib()
    filtered = attr.ib()
    mask = attr.ib()
    frame_number = attr.ib()
    flow = attr.ib(default=None)
    flow_clipped = attr.ib(default=False)
    scaled_thermal = attr.ib(default=None)
    ffc_affected = attr.ib(default=False)
    region = attr.ib(default=None)
    frame_temp_median = attr.ib(default=None)

    def get_channel(self, channel):
        if channel == TrackChannels.thermal:
            return self.thermal
        elif channel == TrackChannels.filtered:
            return self.filtered
        elif channel == TrackChannels.flow:
            return self.flow
        elif channel == TrackChannels.mask:
            return self.mask
        return None

    @classmethod
    def from_channels(
        cls,
        frame,
        channels,
        frame_number,
        flow_clipped=True,
        ffc_affected=False,
        region=None,
    ):
        f = cls(
            None,
            None,
            None,
            frame_number,
            flow_clipped=flow_clipped,
            ffc_affected=ffc_affected,
            region=region,
        )
        flow_h = None
        flow_v = None
        for channel, data in zip(channels, frame):
            if TrackChannels.thermal == channel:
                f.thermal = data
            if TrackChannels.filtered == channel:
                f.filtered = data

            if TrackChannels.mask == channel:
                f.mask = data
            if TrackChannels.flow_h == channel:
                flow_h = data
            if TrackChannels.flow_v == channel:
                flow_v = data
            if TrackChannels.flow == channel:
                f.flow = data
        if flow_h is not None and flow_v is not None:
            flow = np.stack((flow_h, flow_v), axis=2)
            f.flow = flow

        return f

    @classmethod
    def from_array(
        cls,
        frame_arr,
        frame_number,
        flow_clipped=False,
        ffc_affected=False,
        region=None,
    ):
        flow = None
        if len(frame_arr) == 5:
            flow = np.stack(
                (frame_arr[TrackChannels.flow_h], frame_arr[TrackChannels.flow_v]),
                axis=2,
            )
            t = frame_arr[TrackChannels.thermal]
            f = frame_arr[TrackChannels.filtered]
            m = frame_arr[TrackChannels.mask]
        elif len(frame_arr) == 3:
            # no flow so mask is at position 2
            t = frame_arr[TrackChannels.thermal]
            f = frame_arr[TrackChannels.filtered]
            m = frame_arr[2]
        return cls(
            t,
            f,
            m,
            frame_number,
            flow=flow,
            flow_clipped=flow_clipped,
            region=region,
            ffc_affected=ffc_affected,
        )

    def as_array(self, split_flow=True):
        data = [self.thermal]
        if self.filtered is not None:
            data.append(self.filtered)
        else:
            # if no filtered return now as second array must be filtered
            return np.array(data)
        if self.flow is None:
            if self.mask is not None:
                data.append(self.mask)
            return np.asarray(data)
        if split_flow:
            return np.asarray(
                [
                    self.thermal,
                    self.filtered,
                    self.flow[:, :, 0] if self.flow is not None else None,
                    self.flow[:, :, 1] if self.flow is not None else None,
                    self.mask,
                ]
            )
        return np.asarray([self.thermal, self.filtered, self.flow, self.mask])

    def generate_optical_flow(self, opt_flow, prev_frame, flow_threshold=40):
        """
        Generate optical flow from thermal frames
        :param opt_flow: An optical flow algorithm
        """
        height, width = self.thermal.shape
        flow = np.zeros([height, width, 2], dtype=np.float32)
        scaled_thermal = self.thermal.copy()
        scaled_thermal[self.mask == 0] = 0
        scaled_thermal, _ = normalize(scaled_thermal, new_max=255)
        scaled_thermal = np.float32(scaled_thermal)

        # threshold = np.median(self.thermal) + flow_threshold
        # scaled_thermal = np.uint8(np.clip(self.thermal - threshold, 0, 255))
        if prev_frame is not None:
            # for some reason openCV spins up lots of threads for this which really slows things down, so we
            # cap the threads to 2
            cv2.setNumThreads(2)
            flow = opt_flow.calc(prev_frame.scaled_thermal, scaled_thermal, flow)
        self.scaled_thermal = scaled_thermal
        self.flow = flow
        if prev_frame:
            prev_frame.scaled_thermal = None

    def unclip_flow(self):
        if self.flow_clipped:
            self.flow *= 1.0 / 256.0
            self.flow_clipped = False

    def clip_flow(self):
        if self.flow is not None:
            self.flow = get_clipped_flow(self.flow)
            self.flow_clipped = True

    def get_flow_split(self, clip_flow=False):
        if self.flow is not None:
            if self.clip_flow and not self.flow_clipped:
                flow_c = get_clipped_flow(self.flow)
                return flow_c[:, :, 0], flow_c[:, :, 1]

            else:
                return self.flow_h, self.flow_v
        return None, None

    def normalize(self):
        if self.thermal is not None:
            self.thermal, _ = normalize(self.thermal, new_max=255)
        if self.filtered is not None:
            self.filtered, _ = normalize(self.filtered, new_max=255)

    def brightness_adjust(self, adjust):
        if self.thermal is not None:
            self.thermal += adjust

    def contrast_adjust(self, adjust):
        if self.thermal is not None:
            self.thermal *= adjust
        if self.filtered is not None:
            self.filtered *= adjust

    def crop_by_region(self, region, only_thermal=False, out=None):
        # make a new frame cropped by region
        thermal = None
        filtered = None
        mask = None
        flow = None
        if self.thermal is not None:
            thermal = region.subimage(self.thermal)
        if not only_thermal:
            if self.filtered is not None:
                filtered = region.subimage(self.filtered)
            if self.mask is not None:
                mask = region.subimage(self.mask)
            if self.flow is not None:
                flow = region.subimage(self.flow)
        if out:
            out.thermal = thermal
            out.filtered = filtered
            out.mask = mask
            out.flow = flow
            out.region = region
            frame = out
        else:
            frame = Frame(
                thermal,
                filtered,
                mask,
                self.frame_number,
                flow_clipped=self.flow_clipped,
                ffc_affected=self.ffc_affected,
                region=region,
            )
            frame.flow = flow
        return frame

    def resize_with_aspect(self, dim, crop_rectangle, keep_edge=False):
        if self.thermal is not None:
            self.thermal = resize_and_pad(
                self.thermal,
                dim,
                self.region,
                crop_rectangle,
                keep_edge=keep_edge,
            )
        if self.mask is not None:
            self.mask = resize_and_pad(
                self.mask,
                dim,
                self.region,
                crop_rectangle,
                keep_edge=keep_edge,
                pad=0,
                interpolation=cv2.INTER_NEAREST,
            )
        if self.filtered is not None:
            self.filtered = resize_and_pad(
                self.filtered,
                dim,
                self.region,
                crop_rectangle,
                keep_edge=keep_edge,
                pad=0,
            )
        if self.flow is not None:
            flow_h = resize_and_pad(
                self.flow[:, :, 0],
                dim,
                self.region,
                crop_rectangle,
                keep_edge=keep_edge,
                pad=0,
            )
            flow_v = resize_and_pad(
                self.flow[:, :, 1],
                dim,
                self.region,
                crop_rectangle,
                keep_edge=keep_edge,
                pad=0,
            )
            self.flow = np.stack((flow_h, flow_v), axis=2)

    def resize(self, dim):
        self.thermal = resize_cv(self.thermal, dim)
        self.mask = resize_cv(self.mask, dim, interpolation=cv2.INTER_NEAREST)
        if self.flow is not None:
            self.flow = resize_cv(self.flow, dim)
        self.filtered = resize_cv(self.filtered, dim)

    def rotate(self, degrees):
        if self.thermal is not None:
            self.thermal = rotate(self.thermal, degrees)
        if self.mask is not None:
            self.mask = rotate(self.mask, degrees)
        if self.flow is not None:
            self.flow = rotate(self.flow, degrees)
        if self.filtered is not None:
            self.filtered = rotate(self.filtered, degrees)

    def float_arrays(self):
        if self.thermal is not None:
            self.thermal = np.float32(self.thermal)
        if self.mask is not None:
            self.mask = np.float32(self.mask)
        if self.flow is not None:
            self.flow = np.float32(self.flow)
        if self.filtered is not None:
            self.filtered = np.float32(self.filtered)

    def copy(self):
        return Frame(
            None if self.thermal is None else self.thermal.copy(),
            None if self.filtered is None else self.filtered.copy(),
            None if self.mask is None else self.mask.copy(),
            self.frame_number,
            flow=None if self.flow is None else self.flow.copy(),
            flow_clipped=self.flow_clipped,
            ffc_affected=self.ffc_affected,
            region=None if self.region is None else self.region.copy(),
        )

    def flip(self):
        if self.thermal is not None:
            self.thermal = np.flip(self.thermal, axis=1)
        if self.mask is not None:
            self.mask = np.flip(self.mask, axis=1)
        if self.flow is not None:
            self.flow = np.flip(self.flow, axis=1)
        if self.filtered is not None:
            self.filtered = np.flip(self.filtered, axis=1)

    @property
    def flow_h(self):
        if self.flow is None:
            return None
        return self.flow[:, :, 0]

    @property
    def flow_v(self):
        if self.flow is None:
            return None
        return self.flow[:, :, 1]

    @property
    def shape(self):
        return self.thermal.shape



"""
Processes a CPTV file identifying and tracking regions of interest, and saving them in the 'trk' format.
"""

import argparse
import os

from ml_tools.logs import init_logging
from config.config import Config
from mldataset.makedataset import ClipLoader

from pathlib import Path


def parse_params():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--out-folder",
        default="clip_db",
        help="Folder to put processed files",
    )

    parser.add_argument(
        "target",
        default=None,
        help='Target to process, "all" processes all folders, "test" runs test cases, "clean" to remove banned clips from db, or a "cptv" file to run a single source.',
    )
    args = parser.parse_args()
    config = Config.load_from_file(None)
    args.out_folder = Path(args.out_folder)
    return config, args


def main():
    config, args = parse_params()
    init_logging()
    loader = ClipLoader(config)
    loader.process_all(args.target, args.out_folder)


if __name__ == "__main__":
    main()



"""
classifier-pipeline - this is a server side component that manipulates cptv
files and to create a classification model of animals present
Copyright (C) 2018, The Cacophony Project

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
"""

import os.path as path

import attr
import logging

from config import config
from .defaultconfig import DefaultConfig
from ml_tools.previewer import PREVIEW_OPTIONS


@attr.s
class ClassifyConfig(DefaultConfig):
    models = attr.ib()
    meta_to_stdout = attr.ib()
    preview = attr.ib()
    cache_to_disk = attr.ib()
    service_socket = attr.ib()

    @classmethod
    def load(cls, classify):
        return cls(
            models=ClassifyConfig.load_models(classify.get("models")),
            meta_to_stdout=classify["meta_to_stdout"],
            preview=config.parse_options_param(
                "preview", classify["preview"], PREVIEW_OPTIONS
            ),
            cache_to_disk=classify["cache_to_disk"],
            service_socket=classify["service_socket"],
        )

    def load_models(raw):
        if raw is None:
            return None

        models = []
        for model in raw:
            models.append(ModelConfig.load(model))

        return models

    @classmethod
    def get_defaults(cls):
        return cls(
            models=None,
            meta_to_stdout=False,
            preview="none",
            cache_to_disk=False,
            service_socket="/etc/cacophony/classifier",
        )

    def validate(self):
        if self.models is None:
            return
        for model in self.models:
            model.validate()


@attr.s
class ModelConfig:
    DEFAULT_SCORE = 0
    id = attr.ib()
    name = attr.ib()
    type = attr.ib()

    model_file = attr.ib()
    model_weights = attr.ib()

    wallaby = attr.ib()
    tag_scores = attr.ib()
    ignored_tags = attr.ib()
    thumbnail_model = attr.ib()
    reclassify = attr.ib()
    submodel = attr.ib()

    @classmethod
    def load(cls, raw):
        model = cls(
            id=raw["id"],
            name=raw["name"],
            type=raw.get("type", "Keras"),
            model_file=raw["model_file"],
            model_weights=raw.get("model_weights"),
            wallaby=raw.get("wallaby", False),
            tag_scores=load_scores(raw.get("tag_scores", {})),
            ignored_tags=raw.get("ignored_tags", []),
            thumbnail_model=raw.get("thumbnail_model", False),
            reclassify=raw.get("reclassify", None),
            submodel=raw.get("submodel", False),
        )
        return model

    def validate(self):
        if not path.exists(self.model_file):
            logging.warn(f"{self.model_file} does not exist")
            # raise ValueError(f"{self.model_file} does not exist")

    def as_dict(self):
        return attr.asdict(self)


def load_scores(scores):
    scores.setdefault("default", ModelConfig.DEFAULT_SCORE)
    return scores



from abc import ABC, abstractmethod
from threading import Lock
import logging
import numpy as np


class SlidingWindow:
    def __init__(self, shape, dtype):
        self.lock = Lock()
        # if dtype == "O":
        self.frames = [None] * shape
        # else:
        # self.frames = np.empty(shape, dtype)
        self.last_index = None
        self.size = len(self.frames)
        self.oldest_index = None
        self.non_ffc_index = None
        self.ffc = False

    def update_current_frame(self, frame, ffc=False):
        with self.lock:
            if self.last_index is None:
                self.oldest_index = 0
                self.last_index = 0
                if not ffc:
                    self.non_ffc_index = self.oldest_index
            if not ffc and self.ffc:
                self.non_ffc_index = self.last_index

            self.frames[self.last_index] = frame
            self.ffc = ffc

    @property
    def current(self):
        with self.lock:
            if self.last_index is not None:
                return self.frames[self.last_index]
            return None

    def get_frames(self):
        with self.lock:
            if self.last_index is None:
                return []
            frames = []
            cur = self.oldest_index
            end_index = (self.last_index + 1) % self.size
            while len(frames) == 0 or cur != end_index:
                frames.append(self.frames[cur])
                cur = (cur + 1) % self.size
            return frames

    def get(self, i):
        i = i % self.size
        with self.lock:
            return self.frames[i]

    @property
    def oldest_nonffc(self):
        with self.lock:
            if self.non_ffc_index is not None:
                return self.frames[self.non_ffc_index]
            return None

    @property
    def oldest(self):
        with self.lock:
            if self.oldest_index is not None:
                return self.frames[self.oldest_index]
            return None

    def add(self, frame, ffc=False):
        with self.lock:
            if self.last_index is None:
                self.oldest_index = 0
                self.frames[0] = frame
                self.last_index = 0
                if not ffc:
                    self.non_ffc_index = self.oldest_index
            else:
                new_index = (self.last_index + 1) % self.size
                if new_index == self.oldest_index:
                    if self.oldest_index == self.non_ffc_index and not ffc:
                        self.non_ffc_index = (self.oldest_index + 1) % self.size
                    self.oldest_index = (self.oldest_index + 1) % self.size
                self.frames[new_index] = frame
                self.last_index = new_index
            if not ffc and self.ffc:
                self.non_ffc_index = self.last_index
            self.ffc = ffc

    def reset(self):
        with self.lock:
            self.last_index = None
            self.oldest_index = None


class MotionDetector(ABC):
    def __init__(self, thermal_config, headers):
        self.movement_detected = False
        self.use_low_power_mode = thermal_config.recorder.use_low_power_mode
        self.num_frames = 0
        self.rec_window = thermal_config.recorder.rec_window
        self.location_config = thermal_config.location
        self.use_sunrise = self.rec_window.use_sunrise_sunset()
        self.last_sunrise_check = None
        self.location = None
        self.sunrise = None
        self.sunset = None
        self.recording = False

        if self.rec_window.use_sunrise_sunset():
            self.rec_window.set_location(
                *self.location_config.get_lat_long(use_default=True),
                self.location_config.altitude,
            )
        logging.info(
            "Recording window %s - %s ",
            self.rec_window.start.dt,
            self.rec_window.end.dt,
        )
        self.headers = headers

    @property
    def res_x(self):
        return self.headers.res_x

    @property
    def res_y(self):
        return self.headers.res_y

    @abstractmethod
    def process_frame(self, clipped_frame, received_at=None):
        """Tracker type IR or Thermal"""

    @abstractmethod
    def preview_frames(self):
        """Tracker type IR or Thermal"""

    @abstractmethod
    def get_recent_frame(self):
        """Tracker type IR or Thermal"""

    def can_record(self):
        return self.rec_window.inside_window() and not self.use_low_power_mode

    @abstractmethod
    def disconnected(self):
        """Tracker type IR or Thermal"""

    @abstractmethod
    def calibrating(self):
        """Tracker type IR or Thermal"""

    @property
    @abstractmethod
    def background(self):
        """Tracker type IR or Thermal"""


class WeightedBackground:
    def __init__(
        self, edge_pixels, crop_rectangle, res_x, res_y, weight_add, init_average
    ):
        self.edge_pixels = edge_pixels
        self.crop_rectangle = crop_rectangle
        self._background = None
        self.weight_add = weight_add
        self.background_weight = np.zeros(
            (res_y - edge_pixels * 2, res_x - edge_pixels * 2)
        )
        self.average = init_average

    def get_average(self):
        return self.average

    def process_frame(self, frame):
        frame = np.int32(self.crop_rectangle.subimage(frame))
        if self._background is None:
            res_y, res_x = frame.shape
            self._background = np.empty(
                (res_y + self.edge_pixels * 2, res_x + self.edge_pixels * 2)
            )
            self._background[
                self.edge_pixels : res_y + self.edge_pixels,
                self.edge_pixels : res_x + self.edge_pixels,
            ] = frame
            self.average = np.average(frame)
            self.set_background_edges()
            return
        edgeless_back = self.crop_rectangle.subimage(self.background)
        new_background = np.where(
            edgeless_back < frame - self.background_weight,
            edgeless_back,
            frame,
        )
        # update weighting
        # weights could be adjusted to less while recording
        self.background_weight = np.where(
            edgeless_back < frame - self.background_weight,
            self.background_weight + self.weight_add,
            0,
        )
        back_changed = new_background != edgeless_back
        back_changed = np.any(back_changed == True)
        if back_changed:
            edgeless_back[:, :] = new_background
            old_temp = self.average
            self.average = int(round(np.average(edgeless_back)))
            if self.average != old_temp:
                logging.debug(
                    "MotionDetector temp threshold changed from {} to {} ".format(
                        old_temp,
                        self.average,
                    )
                )
            self.set_background_edges()

    def set_background_edges(self):
        for i in range(self.edge_pixels):
            self._background[i] = self._background[self.edge_pixels]
            self._background[-i - 1] = self._background[-self.edge_pixels - 1]
            self._background[:, i] = self._background[:, self.edge_pixels]
            self._background[:, -i - 1] = self._background[:, -1 - self.edge_pixels]

    @property
    def background(self):
        return self._background



"""

Author Giampaolo Ferraro

Date December 2020

Some tools to evaluate a model

"""

import cv2
import argparse
import logging
import pickle
import sys
import os
import time

# os.environ["CUDA_VISIBLE_DEVICES"] = "0"
import json
import pickle
from config.config import Config
from ml_tools.kerasmodel import (
    KerasModel,
    plot_confusion_matrix,
    get_dataset,
    get_excluded,
)
from classify.trackprediction import TrackPrediction

from ml_tools import tools
from ml_tools.trackdatabase import TrackDatabase
from ml_tools.rawdb import RawDatabase
import tensorflow as tf
from ml_tools.dataset import Dataset, filter_clip, filter_track
import pytz
from ml_tools.datasetstructures import SegmentType
from dateutil.parser import parse as parse_date
from sklearn.metrics import confusion_matrix
import matplotlib.pyplot as plt
import numpy as np
from pathlib import Path
from ml_tools.preprocess import preprocess_frame
from ml_tools.frame import Frame
from ml_tools import imageprocessing
import cv2
from config.buildconfig import BuildConfig
from sklearn.metrics import confusion_matrix
from multiprocessing import Pool
from dateutil.parser import parse as parse_date


root_logger = logging.getLogger()
for handler in root_logger.handlers:
    if isinstance(handler, logging.StreamHandler) and handler.stream == sys.stderr:
        root_logger.removeHandler(handler)
crop_rectangle = tools.Rectangle(0, 0, 160, 120)
PROB_THRESHOLD = 0.8

land_birds = [
    "pukeko",
    "california quail",
    "brown quail",
    "black swan",
    "quail",
    "pheasant",
    "penguin",
    "duck",
]


# basic formula to give a number to compare models
def model_score(cm, labels):
    cm = np.around(cm.astype("float") / cm.sum(axis=1)[:, np.newaxis], decimals=2)
    cm = np.nan_to_num(cm)
    fp_index = None
    if "false-positive" in labels:
        fp_index = labels.index("false-positive")
    none_index = None
    unid_index = None
    if "None" in labels:
        none_index = labels.index("None")
    if "unidentified" in labels:
        unid_index = labels.index("unidentified")
    score = 0
    for l_i, l in enumerate(labels):
        fp_acc = 0
        if fp_index is not None:
            fp_acc = cm[l_i][fp_index]
        none_acc = 0
        unid_acc = 0
        accuracy = cm[l_i][l_i]
        if none_index:
            none_acc = cm[l_i][none_index]
        if unid_index:
            unid_acc = cm[l_i][unid_index]
        if l == "bird":
            other_animals = 1 - (fp_acc + none_acc + unid_acc)
            score += accuracy * 1.2 - other_animals
        elif l in ["vehicle", "wallaby"]:
            score += accuracy * 0.8
        elif l in ["mustelid", "human"]:
            score += accuracy * 0.9
        elif l not in ["None", "unidentified"]:
            score += accuracy * 1
    logging.info("Model accuracy score is %s", score)


def get_mappings(label_paths):
    regroup = {}
    for l, path in label_paths.items():
        if l in land_birds:
            regroup[l] = l
            continue
        split_path = path.split(".")
        if len(split_path) == 1:
            regroup[l] = l
        elif path.startswith("all.mammal"):
            if len(split_path) == 4:
                regroup[l] = split_path[-2]
            else:
                regroup[l] = l
        else:
            # print("l", l, " has ", path)
            parent = split_path[-2]
            # print("Parent is", parent, path)
            if parent == "kiwi" or split_path[-1] == "kiwi":
                regroup[l] = "kiwi"
            elif parent == "other":
                regroup[l] = l

            else:
                if "bird." in path:
                    regroup[l] = "bird"

                elif len(split_path) > 2:
                    regroup[l] = split_path[-3]
                else:
                    regroup[l] = split_path[-1]

    return regroup


def load_args():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-m",
        "--model-file",
        help="Path to model file to use, will override config model",
    )
    parser.add_argument("-w", "--weights", help="Weights to load into model")

    parser.add_argument(
        "-s",
        "--dataset",
        default="test",
        help="Dataset to use train, validation, test ( Default)",
    )

    parser.add_argument(
        "--evaluate-dir",
        help="Evalute directory of cptv files",
    )
    parser.add_argument(
        "--model-metadata",
        help="Meta data file for model, used with confusion from meta",
    )
    parser.add_argument("-c", "--config-file", help="Path to config file to use")

    parser.add_argument("-d", "--date", help="Use clips after this")

    parser.add_argument("--split-file", help="Use split for evaluation")
    parser.add_argument(
        "--confusion-from-meta",
        action="count",
        help="Use metadata to produce a confusion matrix",
    )

    parser.add_argument(
        "confusion",
        help="Confusion matrix filename, used if you want to save confusion matrix image",
    )
    parser.add_argument(
        "--threshold",
        default=0.5,
        type=float,
        help="Prediction threshold default 0.5",
    )
    args = parser.parse_args()
    if args.date:
        args.date = parse_date(args.date)
        args.date = args.date.replace(tzinfo=pytz.UTC)
    return args


def init_logging(timestamps=False):
    """Set up logging for use by various classifier pipeline scripts.

    Logs will go to stderr.
    """

    fmt = "%(levelname)7s %(message)s"
    if timestamps:
        fmt = "%(asctime)s " + fmt
    logging.basicConfig(
        stream=sys.stderr, level=logging.INFO, format=fmt, datefmt="%Y-%m-%d %H:%M:%S"
    )


def filter_diffs(track_frames, background):
    max_diff = 0
    min_diff = 0
    for f in track_frames:
        r = f.region
        if r.blank or r.width <= 0 or r.height <= 0:
            continue

        diff_frame = np.float32(f.thermal) - r.subimage(background)
        new_max = np.amax(diff_frame)
        new_min = np.amin(diff_frame)
        if new_min < min_diff:
            min_diff = new_min
        if new_max > max_diff:
            max_diff = new_max
    return min_diff, max_diff


# evaluate a confusion matrix from metadata of files, already evaluated by our current model on browse


def metadata_confusion(dir, confusion_file, after_date=None, model_metadata=None):
    with open("label_paths.json", "r") as f:
        label_paths = json.load(f)
    label_mapping = get_mappings(label_paths)
    if model_metadata is not None and Path(model_metadata).exists():
        with open(model_metadata, "r") as t:
            # add in some metadata stats
            model_meta = json.load(t)
        labels = model_meta.get("labels", [])
        excluded_labels = model_meta.get("excluded_labels", {})
        remapped_labels = model_meta.get("remapped_labels", {})
        for k, v in remapped_labels.items():
            if v == "land-bird":
                remapped_labels[k] = "bird"
        if "None" not in labels:
            labels.append("None")
        if "unidentified" not in labels:
            labels.append("unidentified")
    else:
        labels = [
            "bird",
            "cat",
            "deer",
            "dog",
            "falsepositive",
            "hedgehog",
            "human",
            "kiwi",
            "leporidae",
            "mustelid",
            "penguin",
            "possum",
            "rodent",
            "sheep",
            "vehicle",
            "wallaby",
            "landbird",
            "None",
            "unidentified",
        ]
        excluded_labels, remapped_labels = get_excluded("thermal")
    logging.info(
        "Labels are %s excluded %s remapped %s",
        labels,
        excluded_labels,
        remapped_labels,
    )
    y_true = []
    y_pred = []
    dir = Path(dir)
    for cptv_file in dir.glob(f"**/*cptv"):
        meta_f = cptv_file.with_suffix(".txt")
        if not meta_f.exists():
            continue
        meta_data = None
        with open(meta_f, "r") as t:
            # add in some metadata stats
            meta_data = json.load(t)
        rec_time = parse_date(meta_data["recordingDateTime"])
        if after_date is not None and rec_time <= after_date:
            continue
        for track in meta_data.get("Tracks", []):
            tags = track.get("tags", [])
            human_tags = [
                tag.get("what") for tag in tags if tag.get("automatic") == False
            ]
            human_tags = set(human_tags)
            if len(human_tags) > 1:
                print("Conflicting tags for ", track.get("id"), cptv_file)
            if len(human_tags) == 0:
                print("No humans in ", meta_f)
                continue
            human_tag = human_tags.pop()
            human_tag = label_mapping.get(human_tag, human_tag)
            if human_tag in excluded_labels:
                logging.info("Excluding %s", human_tag)
                continue
            if human_tag in remapped_labels:
                logging.info(
                    "Remapping %s to %s", human_tag, remapped_labels[human_tag]
                )
                human_tag = remapped_labels[human_tag]
            # if human_tag not in labels:
            # logging.info("Excluding %s", human_tag)

            ai_tags = []
            for tag in tags:
                if tag.get("automatic") is True:
                    data = tag.get("data", {})
                    if isinstance(data, str):
                        if data == "Master":
                            ai_tags.append(tag["what"])
                    elif data.get("name") == "Master":
                        ai_tags.append(tag["what"])

            y_true.append(human_tag)
            if human_tag not in labels:
                labels.append(human_tag)
            if len(ai_tags) == 0:
                y_pred.append("None")
            else:
                y_pred.append(ai_tags[0])
                if ai_tags[0] not in labels:
                    labels.append(ai_tags[0])

    cm = confusion_matrix(y_true, y_pred, labels=labels)
    # Log the confusion matrix as an image summary.
    figure = plot_confusion_matrix(cm, class_names=labels)
    plt.savefig(confusion_file, format="png")
    # cm = np.around(cm.astype("float") / cm.sum(axis=1)[:, np.newaxis], decimals=2)
    # cm = np.nan_to_num(cm)
    model_score(cm, labels)


EXCLUDED_TAGS = ["poor tracking", "part", "untagged", "unidentified"]
worker_model = None
after_date = None


def init_worker(model, date):
    global worker_model, after_date
    worker_model = model
    after_date = date


def load_clip_data(cptv_file):
    # for clip in dataset.clips:
    reason = {}
    clip_db = RawDatabase(cptv_file)
    clip = clip_db.get_clip_tracks(BuildConfig.DEFAULT_GROUPS)
    if clip is None:
        logging.warn("No clip for %s", cptv_file)
        return None

    if filter_clip(clip, None, None, reason, after_date=after_date):
        # logging.info("Filtering %s", cptv_file)
        return None
    clip.tracks = [
        track for track in clip.tracks if not filter_track(track, EXCLUDED_TAGS, reason)
    ]
    if len(clip.tracks) == 0:
        logging.info("No tracks after filtering %s", cptv_file)
        return None
    clip_db.load_frames()
    segment_frame_spacing = int(round(clip.frames_per_second))
    thermal_medians = []
    for f in clip_db.frames:
        thermal_medians.append(np.median(f.thermal))
    thermal_medians = np.uint16(thermal_medians)
    data = []
    for track in clip.tracks:
        try:
            frames, preprocessed, masses = worker_model.preprocess(
                clip_db, track, frames_per_classify=25, dont_filter=True, min_segments=1
            )
            data.append(
                (
                    f"{track.clip_id}-{track.get_id()}",
                    track.label,
                    frames,
                    preprocessed,
                    masses,
                )
            )
        except:
            logging.error("Could not load %s", clip.clip_id, exc_info=True)
    return data


def load_split_file(split_file):
    with open(split_file, "r") as f:
        split = json.load(f)
    return split


def evaluate_dir(
    model,
    dir,
    config,
    confusion_file,
    split_file=None,
    split_dataset="test",
    threshold=0.5,
    after_date=None,
):
    logging.info("Evaluating cptv files in %s with threshold %s", dir, threshold)

    with open("label_paths.json", "r") as f:
        label_paths = json.load(f)
    label_mapping = get_mappings(label_paths)
    reason = {}
    y_true = []
    y_pred = []
    if split_file is not None:
        split_json = load_split_file(split_file)
        files = split_json.get(split_dataset)
        files = [dir / f["source"] for f in files]
        logging.info(
            "Splitting on %s dataset %s files %s ...",
            split_file,
            split_dataset,
            files[:2],
        )
    else:
        files = list(dir.glob(f"**/*cptv"))
    files.sort()
    # files = files[:8]
    start = time.time()
    # quite faster with just one process for loading and using main process for predicting
    with Pool(
        processes=1,
        initializer=init_worker,
        initargs=(
            model,
            after_date,
        ),
    ) as pool:
        for clip_data in pool.imap_unordered(load_clip_data, files):
            if clip_data is None:
                continue
            for data in clip_data:
                label = data[1]
                preprocessed = data[3]
                if len(preprocessed) == 0:
                    logging.info("No data found for %s", data[0])
                    y_true.append(label_mapping.get(label, label))
                    y_pred.append("None")
                    continue
                output = model.predict(preprocessed)

                prediction = TrackPrediction(data[0], model.labels)
                masses = np.array(data[4])
                masses = masses[:, None]
                top_score = None
                if model.params.multi_label is True:
                    #     # every label could be 1 for each prediction
                    top_score = np.sum(masses)
                #     smoothed = output
                # else:
                smoothed = output * masses
                prediction.classified_clip(
                    output, smoothed, data[2], masses, top_score=top_score
                )
                y_true.append(label_mapping.get(label, label))
                predicted_labels = [prediction.predicted_tag()]
                confidence = prediction.max_score
                predicted_tag = "None"
                if confidence < threshold:
                    y_pred.append("unidentified")
                elif len(predicted_labels) == 0:
                    y_pred.append("None")
                else:
                    logging.info("Predicted  %s", predicted_labels)
                    predicted_tag = ",".join(predicted_labels)
                    y_pred.append(predicted_tag)
                if y_pred[-1] != y_true[-1]:
                    logging.info(
                        "%s predicted %s but should be %s with confidence %s",
                        data[0],
                        y_pred[-1],
                        label,
                        np.round(100 * prediction.class_best_score),
                    )
    model.labels.append("None")
    model.labels.append("unidentified")
    cm = confusion_matrix(y_true, y_pred, labels=model.labels)
    npy_file = Path(confusion_file).with_suffix(".npy")
    logging.info("Saving %s", npy_file)
    np.save(str(npy_file), cm)

    # Log the confusion matrix as an image summary.
    figure = plot_confusion_matrix(cm, class_names=model.labels)
    plt.savefig(confusion_file, format="png")
    logging.info("Saving %s", Path(confusion_file).with_suffix(".png"))

    model_score(cm, model.labels)


min_tag_clarity = 0.2
min_tag_confidence = 0.8


# was used to save model architecture differently so that the predictions were the same
def re_save(model_file, weights, config):
    tf.random.set_seed(1)
    model = KerasModel(train_config=config.train)
    model.load_model(model_file.parent)
    model.model.load_weights(weights).expect_partial()
    model = model.model

    output = model.layers[1].output
    for new_l in model.layers[2:]:
        print(new_l.name)
        output = new_l(output)
    new_model = tf.keras.models.Model(model.layers[1].input, outputs=output)

    print("Saving", model_file.parent / "re_save")
    new_model.summary()
    new_model.save(model_file.parent / "re_save")


def main():
    args = load_args()
    init_logging()
    config = Config.load_from_file(args.config_file)
    print("Loading config", args.config_file)
    weights = None
    if args.model_file:
        model_file = Path(args.model_file)
    if args.weights:
        weights = model_file / args.weights
    base_dir = Path(config.base_folder) / "training-data"
    if args.evaluate_dir and args.confusion_from_meta:
        metadata_confusion(
            Path(args.evaluate_dir), args.confusion, args.date, args.model_metadata
        )
    else:

        model = KerasModel(train_config=config.train)
        model.load_model(model_file, training=False, weights=weights)
        if args.evaluate_dir:
            evaluate_dir(
                model,
                Path(args.evaluate_dir),
                config,
                args.confusion,
                args.split_file,
                args.dataset,
                threshold=args.threshold,
                after_date=args.date,
            )
        elif args.dataset:
            model_labels = model.labels.copy()
            model.load_training_meta(base_dir)
            # model.labels = model_labels
            if model.params.multi_label:
                model.labels.append("land-bird")
            excluded, remapped = get_excluded(model.data_type)

            if model.params.excluded_labels is not None:
                excluded = model.params.excluded_labels

            if model.params.remapped_labels is not None:
                remapped = model.params.remapped_labels

            files = base_dir / args.dataset
            dataset, _, new_labels, _ = get_dataset(
                files,
                model.data_type,
                model.labels,
                model_labels=model_labels,
                batch_size=64,
                image_size=model.params.output_dim[:2],
                preprocess_fn=model.preprocess_fn,
                augment=False,
                resample=False,
                include_features=model.params.mvm,
                one_hot=True,
                deterministic=True,
                shuffle=False,
                excluded_labels=excluded,
                remapped_labels=remapped,
                multi_label=model.params.multi_label,
                include_track=True,
                cache=True,
                channels=model.params.channels,
                num_frames=model.params.square_width**2,
            )
            model.labels = new_labels
            logging.info(
                "Dataset loaded %s, using labels %s",
                args.dataset,
                model.labels,
            )
            model.confusion_tracks(dataset, args.confusion, threshold=args.threshold)


if __name__ == "__main__":
    main()



import os
import argparse

from ml_tools.trackdatabase import TrackDatabase
from config.config import Config
from datetime import timedelta
from datetime import date
from pathlib import Path
from dateutil.parser import parse as parse_date

parser = argparse.ArgumentParser()
parser.add_argument("data_dir", help="Directory of cptv files")
args = parser.parse_args()
args.data_dir = Path(args.data_dir)
latest_date = None
for db_clip in args.data_dir.glob(f"**/*.cptv"):
    file_name = db_clip.name
    hyphen = file_name.index("-")
    date_s = file_name[hyphen + 1 : hyphen + 16]
    cptv_dt = parse_date(date_s)
    if latest_date is None or cptv_dt > latest_date:
        latest_date = cptv_dt


month_ago = latest_date - timedelta(days=30 * 6)
month_ago = month_ago.strftime("%Y-%m-%d 00:00:00")
print(month_ago)



# Script to extra images from datasets with meta_data #
# This can be used to give to people who dont want to work with the h5py db
from multiprocessing import Process, Queue
from PIL import Image
import numpy as np
import pickle
import argparse
import logging
import os
import sys
from config.config import Config
from ml_tools.imageprocessing import normalize
import pickle


def save_job(queue, dataset, folder, labels_dir):
    i = 0
    while True:
        i += 1
        track = queue.get()
        if track == "DONE":
            break
        else:
            save_track(dataset, track, folder, labels_dir)
        if i % 50 == 0:
            logging.info("%s jobs left", queue.qsize())


def save_track(dataset, track, folder, labels_dir, ext="png"):
    track_data = dataset.fetch_track(track, original=True)
    clip_meta = dataset.db.get_clip_meta(track.clip_id)

    clip_dir = os.path.join(folder, str(track.clip_id))
    if not os.path.isdir(clip_dir):
        logging.debug("Creating %s", clip_dir)
        os.mkdir(clip_dir)

    background = dataset.db.get_clip_background(track.clip_id)
    if background is not None:
        background, _ = normalize(background, new_max=255)

        img = Image.fromarray(np.uint8(background))
        filename = "{}-background.{}".format(track.clip_id, ext)
        logging.debug("Saving %s", os.path.join(clip_dir, filename))
        img.save(os.path.join(clip_dir, filename))

    for i, frame in enumerate(track_data):
        thermal = frame.thermal
        normed, _ = normalize(thermal, new_max=255)
        img = Image.fromarray(np.uint8(normed))
        filename = "{}-{}.{}".format(track.clip_id, i + track.start_frame, ext)
        logging.debug("Saving %s", os.path.join(clip_dir, filename))

        img.save(os.path.join(clip_dir, filename))
    save_metadata(clip_meta, track, labels_dir)


def save_metadata(clip_meta, track, folder):
    meta_dir = os.path.join(folder, track.label)
    if not os.path.isdir(meta_dir):
        logging.debug("Creating %s", meta_dir)
        os.mkdir(meta_dir)
    filename = "{}-{}.txt".format(track.clip_id, track.track_id)
    fullpath = os.path.join(meta_dir, filename)
    with open(fullpath, "w") as f:
        f.write(track.toJSON(clip_meta))


def save_all(dataset, worker_threads, folder):
    if not os.path.isdir(folder):
        logging.debug("Creating %s", folder)
        os.mkdir(folder)
    labels_dir = os.path.join(folder + "-labels")
    if not os.path.isdir(labels_dir):
        logging.debug("Creating %s", labels_dir)
        os.mkdir(labels_dir)
    job_queue = Queue()
    processes = []
    for i in range(max(1, 4)):
        p = Process(
            target=save_job,
            args=(job_queue, dataset, folder, labels_dir),
        )
        processes.append(p)
        p.start()

    for track in dataset.tracks:
        job_queue.put(track)

    logging.debug("Processing %d", len(dataset.tracks))
    for i in range(len(processes)):
        job_queue.put("DONE")
    for process in processes:
        process.join()


def load_args():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-f", "--base-folder", help="Base folder to save, train test and val"
    )

    parser.add_argument("-c", "--config-file", help="Path to config file to use")
    args = parser.parse_args()
    return args


def init_logging(timestamps=False):
    """Set up logging for use by various classifier pipeline scripts.

    Logs will go to stderr.
    """

    fmt = "%(levelname)7s %(message)s"
    if timestamps:
        fmt = "%(asctime)s " + fmt
    logging.basicConfig(
        stream=sys.stderr, level=logging.INFO, format=fmt, datefmt="%Y-%m-%d %H:%M:%S"
    )


args = load_args()
init_logging()
config = Config.load_from_file(args.config_file)
datasets = ["train", "validation", "test"]

base_dir = "."
if args.base_folder:
    base_dir = args.base_folder
if not os.path.isdir(base_dir):
    logging.debug("Creating %s", base_dir)
    os.mkdir(base_dir)
for i, name in enumerate(datasets):
    dataset = pickle.load(open(f"{os.path.join(config.tracks_folder, name)}.dat", "rb"))
    dataset.load_db()
    save_all(dataset, config.worker_threads, os.path.join(base_dir, dataset.name))



"""
classifier-pipeline - this is a server side component that manipulates cptv
files and to create a classification model of animals present
Copyright (C) 2018, The Cacophony Project

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
"""

import attr

from .defaultconfig import (
    DefaultConfig,
    deep_copy_map_if_key_not_exist,
)
from .trackingmotionconfig import TrackingMotionConfig
from track.track import RegionTracker


@attr.s
class TrackingConfig(DefaultConfig):
    tracker = attr.ib()
    params = attr.ib()
    type = attr.ib()
    motion = attr.ib()
    edge_pixels = attr.ib()
    # dilation_pixels = attr.ib()
    min_dimension = attr.ib()
    frame_padding = attr.ib()
    track_smoothing = attr.ib()
    denoise = attr.ib()

    high_quality_optical_flow = attr.ib()
    flow_threshold = attr.ib()
    max_tracks = attr.ib()
    track_overlap_ratio = attr.ib()
    min_duration_secs = attr.ib()
    track_min_offset = attr.ib()
    track_min_mass = attr.ib()
    aoi_min_mass = attr.ib()
    aoi_pixel_variance = attr.ib()
    cropped_regions_strategy = attr.ib()
    enable_track_output = attr.ib()
    min_tag_confidence = attr.ib()
    moving_vel_thresh = attr.ib()
    min_moving_frames = attr.ib()
    max_blank_percent = attr.ib()
    max_mass_std_percent = attr.ib()

    max_jitter = attr.ib()
    # used to provide defaults
    filters = attr.ib()
    areas_of_interest = attr.ib()
    # filter regions out by mass and variance before matching to a track
    filter_regions_pre_match = attr.ib()
    min_hist_diff = attr.ib()

    @classmethod
    def load(cls, tracking):
        if tracking is None:
            return None
        trackers = {}
        for type, raw_tracker in tracking.items():
            if raw_tracker is None:
                raw_tracker = {}
            tracker = TrackingConfig.load_type(raw_tracker, type)
            trackers[tracker.type] = tracker
        return trackers

    @classmethod
    def load_type(cls, tracking, type):
        defaults = cls.get_type_defaults(type)
        deep_copy_map_if_key_not_exist(defaults.as_dict(), tracking)
        return cls(
            tracker=tracking["tracker"],
            params=tracking["params"],
            type=type,
            motion=TrackingMotionConfig.load(tracking.get("motion")),
            min_dimension=tracking["min_dimension"],
            edge_pixels=tracking["edge_pixels"],
            # dilation_pixels=tracking["dilation_pixels"],
            frame_padding=tracking["frame_padding"],
            track_smoothing=tracking["track_smoothing"],
            denoise=tracking["denoise"],
            high_quality_optical_flow=tracking["high_quality_optical_flow"],
            flow_threshold=tracking["flow_threshold"],
            max_tracks=tracking["max_tracks"],
            moving_vel_thresh=tracking["filters"]["moving_vel_thresh"],
            track_overlap_ratio=tracking["filters"]["track_overlap_ratio"],
            min_duration_secs=tracking["filters"]["min_duration_secs"],
            track_min_offset=tracking["filters"]["track_min_offset"],
            track_min_mass=tracking["filters"]["track_min_mass"],
            cropped_regions_strategy=tracking["areas_of_interest"][
                "cropped_regions_strategy"
            ],
            aoi_min_mass=tracking["areas_of_interest"]["min_mass"],
            aoi_pixel_variance=tracking["areas_of_interest"]["pixel_variance"],
            enable_track_output=tracking["enable_track_output"],
            min_tag_confidence=tracking["min_tag_confidence"],
            min_moving_frames=tracking["min_moving_frames"],
            max_blank_percent=tracking["max_blank_percent"],
            max_jitter=tracking["max_jitter"],
            filters=tracking["filters"],
            areas_of_interest=tracking["areas_of_interest"],
            max_mass_std_percent=tracking["max_mass_std_percent"],
            filter_regions_pre_match=tracking["filter_regions_pre_match"],
            min_hist_diff=tracking["min_hist_diff"],
        )

    @classmethod
    def get_defaults(cls):
        default_tracking = {}
        default_tracking["thermal"] = cls.get_type_defaults("thermal")
        default_tracking["IR"] = cls.get_type_defaults("IR")
        return default_tracking

    @classmethod
    def get_type_defaults(cls, type):
        default_tracking = cls(
            motion=TrackingMotionConfig.get_defaults(),
            edge_pixels=1,
            frame_padding=4,
            min_dimension=0,
            # dilation_pixels=2,
            track_smoothing=False,
            denoise=True,
            high_quality_optical_flow=False,
            flow_threshold=40,
            max_tracks=None,
            filters={
                "track_overlap_ratio": 0.5,
                "min_duration_secs": 0,
                "track_min_offset": 4.0,
                "track_min_mass": 2.0,
                "moving_vel_thresh": 4,
            },
            areas_of_interest={
                "min_mass": 4.0,
                "pixel_variance": 2.0,
                "cropped_regions_strategy": "cautious",
            },
            # defaults provided in dictionaries, placesholders to stop init complaining
            aoi_min_mass=4.0,
            aoi_pixel_variance=2.0,
            cropped_regions_strategy="cautious",
            track_min_offset=4.0,
            track_min_mass=2.0,
            track_overlap_ratio=0.5,
            min_duration_secs=0,
            min_tag_confidence=0.8,
            enable_track_output=True,
            moving_vel_thresh=4,
            min_moving_frames=2,
            max_blank_percent=30,
            max_mass_std_percent=RegionTracker.MASS_CHANGE_PERCENT,
            max_jitter=20,
            tracker="RegionTracker",
            type="thermal",
            params={
                "base_distance_change": 450,
                "min_mass_change": 20,
                "restrict_mass_after": 1.5,
                "mass_change_percent": 0.55,
                "max_distance": 2000,
                "max_blanks": 18,
                "velocity_multiplier": 2,
                "base_velocity": 2,
            },
            filter_regions_pre_match=True,
            min_hist_diff=None,
        )
        if type == "IR":
            # default_tracking.min_hist_diff = 0.95
            default_tracking.filters["min_duration_secs"] = 0
            default_tracking.min_duration_secs = 0
            default_tracking.filter_regions_pre_match = False
            default_tracking.areas_of_interest["pixel_variance"] = 0
            default_tracking.areas_of_interest["min_mass"] = 0
            default_tracking.filters["track_min_offset"] = 7
            default_tracking.track_min_offset = 20
            default_tracking.min_dimension = 10
            default_tracking.min_tracks = None
            default_tracking.frame_padding = 10
            default_tracking.edge_pixels = 0
            default_tracking.tracker = "RegionTracker"
            default_tracking.type = "IR"
            default_tracking.params = {
                "base_distance_change": 12000,
                "min_mass_change": None,
                "restrict_mass_after": 1.5,
                "mass_change_percent": None,
                "max_distance": 30752,
                "max_blanks": 18,
                "velocity_multiplier": 8,
                "base_velocity": 10,
            }
        return default_tracking

    #
    # def load_trackers(raw):
    #     if raw is None:
    #         return None
    #     trackers = {}
    #     for raw_tracker in raw.values():
    #         tracker = TrackingConfig.load(raw_tracker)
    #         trackers[tracker.type] = tracker
    #     return trackers

    def validate(self):
        return True

    def as_dict(self):
        return attr.asdict(self)

    def rescale(self, scale):
        # adjust numbers if we rescale frame sizes
        self.frame_padding = int(scale * self.frame_padding)
        self.min_dimension = int(scale * self.min_dimension)
        if self.params["base_distance_change"]:
            self.params["base_distance_change"] *= scale
        if self.params["min_mass_change"]:
            self.params["min_mass_change"] *= scale
        if self.params["max_distance"]:
            self.params["max_distance"] *= scale
        if self.params["base_velocity"]:
            self.params["base_velocity"] *= scale
        self.track_min_offset *= scale
        self.track_min_mass *= scale
        self.aoi_min_mass *= scale


#
#
# @attr.s
# class TrackerConfig(DefaultConfig):
#
#     tracker = attr.ib()
#     params = attr.ib()
#     type = attr.ib()
#
#     @classmethod
#     def load(cls, raw):
#         defaults = cls.get_defaults()
#         deep_copy_map_if_key_not_exist(defaults.as_dict(), raw)
#
#         return cls(
#             tracker=raw["tracker"],
#             params=raw["params"],
#             type=raw["type"],
#         )
#
#     def as_dict(self):
#         return attr.asdict(self)
#
#     @classmethod
#     def get_defaults(cls):
#         return cls(
#             tracker="RegionTracker",
#             type="IR",
#             params={
#                 "base_distance_change": 11250,
#                 "min_mass_change": 20 * 4,
#                 "restrict_mass_after": 1.5,
#                 "mass_change_percent": 0.55,
#                 "max_distance": 30752,
#             },
#         )
#
#     def validate(self):
#         return True



import locale
import os
import subprocess
import logging


class MPEGCreator:
    """
    This class allows an MPEG video to be created frame by frame.

    Usage:

        m = MPEGCreator("out.mp4")
        m.next_frame(frame0)
        m.next_frame(frame1)
        ...
        m.close()

    The output from ffmpeg is available via the `output` property.
    """

    def __init__(
        self,
        filename,
        quality=21,
        fps=9,
        codec="libx264",
        bitrate="400K",
        pix_fmt="bgr24",
    ):
        self.filename = filename
        self.quality = quality
        self._ffmpeg = None
        self._output = []
        self.fps = fps
        self.codec = codec
        self.bitrate = bitrate
        self.pix_fmt = pix_fmt

    def next_frame(self, frame, shape=None):
        if self._ffmpeg is None:
            if shape is None:
                height, width, _ = frame.shape
            else:
                height, width, _ = shape
            self._ffmpeg = self._start(width, height)
        self._ffmpeg.stdin.write(frame.tobytes())

    def close(self):
        if not self._ffmpeg:
            return
        self._ffmpeg.stdin.close()

        return_code = self._ffmpeg.wait(timeout=60)
        if return_code:
            self._collect_output()
            raise Exception(
                "ffmpeg failed with error {}. output:\n{}".format(
                    return_code, self.output
                )
            )

    @property
    def output(self):
        encoding = locale.getpreferredencoding(False)
        return (b"".join(self._output)).decode(encoding)

    def _start(self, width, height):
        command = self.get_ffmpeg_command(width, height)
        logging.debug("command is %s", " ".join(command))
        proc = subprocess.Popen(
            command,
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            bufsize=8 * 1024,
        )
        return proc

    def _collect_output(self):
        buf = self._ffmpeg.stdout.read()
        if buf:
            self._output.append(buf)

    def get_ffmpeg_command(self, width, height, quality=18):
        if os.name == "nt":
            FFMPEG_BIN = "ffmpeg.exe"  # on Windows
        else:
            FFMPEG_BIN = "ffmpeg"  # on Linux ans Mac OS

        command = [
            FFMPEG_BIN,
            "-y",  # overwrite output file if it exists
            "-f",
            "rawvideo",
            "-vcodec",
            "rawvideo",
            "-loglevel",
            "warning",  # minimal output
            "-s",
            str(width) + "x" + str(height),  # size of one frame
            "-pix_fmt",
            self.pix_fmt,
            "-r",
            str(self.fps),  # frames per second
            "-i",
            "-",  # The imput comes from a pipe
            "-an",  # Tells FFMPEG not to expect any audio
            "-c:v",
            self.codec,
            "-b:v",
            "1000k",
            "-preset",
            "veryfast",
            str(self.filename),
        ]
        return command



class Telemetry:
    def __init__(self):
        self.telemetry_revision = None
        self.time_on = None
        self.status_bits = None
        self.reserved5 = None
        self.software_revision = None
        self.reserved17 = None
        self.frame_counter = None
        self.frame_mean = None
        self.fpa_temp_counts = None
        self.fpa_temp = None
        self.reserved25 = None
        self.fpa_temp_last_ffc = None
        self.last_ffc_time = None
        self.ffc_imminent = False
        self.ffc_status = None



"""

Author GP

Date 2023


"""

import os
import logging
import filelock
import numpy as np
from dateutil.parser import parse as parse_date
from .frame import Frame, TrackChannels
import json
from pathlib import Path
import numpy as np
from track.region import Region
from ml_tools.datasetstructures import TrackHeader, ClipHeader
from track.track import Track
from track.cliptrackextractor import is_affected_by_ffc
from cptv_rs_python_bindings import CptvReader
from ml_tools.rectangle import Rectangle
from config.buildconfig import BuildConfig
from datetime import timedelta
from piclassifier.motiondetector import WeightedBackground

special_datasets = [
    "tag_frames",
    "original_frames",
    "background_frame",
    "predictions",
    "overlay",
]

FPS = 9

RES_X = 160
RES_Y = 120


class RawDatabase:
    def __init__(self, database_filename):
        self.file = Path(database_filename)
        self.meta_data_file = self.file.with_suffix(".txt")
        self._meta_data = None
        self.background = None
        self.ffc_frames = None
        self.frames = None
        self.crop_rectangle = Rectangle(1, 1, 160 - 2, 120 - 2)

    def frames_kept(self):
        return None

    def get_frame(self, frame_number):
        if self.frames is None or frame_number > len(self.frames):
            return None
        return self.frames[frame_number]

    def get_frames(self):
        return self.frames

    def get_clip_background(self):
        return self.background

    def load_frames(self):
        ffc_frames = []
        cptv_frames = []
        background = None
        tracker_version = self.meta_data.get("tracker_version")
        frame_i = 0
        reader = CptvReader(str(self.file))
        header = reader.get_header()

        background_alg = None
        back_processed = False
        self.frames = []
        while True:
            frame = reader.next_frame()
            if frame is None:
                break

            if background_alg is None:
                average = np.mean(frame.pix)
                if average > 10000:
                    model = "lepton3.5"
                    weight_add = 1
                else:
                    model = "lepton3"
                    weight_add = 0.1
                background_alg = WeightedBackground(
                    self.crop_rectangle.x,
                    self.crop_rectangle,
                    RES_X,
                    RES_Y,
                    weight_add,
                    average,
                )
                background_alg.process_frame(frame.pix)
                back_processed = True
                background = background_alg.background

            if frame.background_frame:
                # background = frame.pix
                # bug in previous tracker version where background was first frame
                if tracker_version >= 10:
                    continue
            ffc = is_affected_by_ffc(frame)
            if ffc:
                ffc_frames.append(frame_i)
            self.frames.append(
                Frame(
                    frame.pix,
                    np.float32(frame.pix) - background_alg.background,
                    None,
                    frame_i,
                )
            )

            if not back_processed:
                background_alg.process_frame(frame.pix)

            frame_i += 1

        self.ffc_frames = ffc_frames
        self.background = background

    @property
    def meta_data(self):
        if self._meta_data is not None:
            return self._meta_data
        if not self.meta_data_file.is_file():
            logging.warn("Could not load meta data for %s", self.meta_data_file)
            return None
        with open(self.meta_data_file, "r") as t:
            # add in some metadata stats
            self._meta_data = json.load(t)
        return self._meta_data

    def get_clip_tracks(self, tag_precedence):
        metadata = self.meta_data
        if metadata is None:
            return None
        edge_pixels = metadata.get("edgePixels", 1)
        resx = metadata.get("resX", 160)
        resy = metadata.get("resY", 140)

        self.crop_rectangle = Rectangle(
            edge_pixels, edge_pixels, resx - edge_pixels * 2, resy - edge_pixels * 2
        )
        location = metadata.get("location")
        lat = None
        lng = None
        country_code = None
        if location is not None:
            try:
                lat = location.get("lat")
                lng = location.get("lng")
                if lat is not None and lng is not None:
                    for country, location in BuildConfig.COUNTRY_LOCATIONS.items():
                        if location.contains(lng, lat):
                            country_code = country
                            break
            except:
                logging.error("Could not parse lat lng", exc_info=True)
                pass

        clip_header = ClipHeader(
            clip_id=int(metadata["id"]),
            station_id=metadata.get("stationId"),
            source_file=self.file,
            location=None if lat is None or lng is None else (lng, lat),
            camera=metadata.get("deviceId"),
            rec_time=parse_date(metadata["recordingDateTime"]),
            frames_per_second=10 if self.file.suffix == "mp4" else 9,
            events=metadata.get("event", ""),
            trap=metadata.get("trap", ""),
            tracks=[],
            ffc_frames=self.ffc_frames,
            country_code=country_code,
        )
        tracks = metadata.get("Tracks", [])
        fp_labels = metadata.get("fp_model_labels")
        fp_index = None
        if fp_labels is not None:
            fp_index = fp_labels.index("false-positive")
        meta = []
        for track_meta in tracks:
            tags = track_meta.get("tags", [])
            tag = Track.get_best_human_tag(tags, tag_precedence, 0)
            human_tag = None
            human_tag_confidence = None
            master_tag = [
                t
                for t in tags
                if t.get("automatic")
                and not isinstance(t.get("data", ""), str)
                and t.get("data", {}).get("name") == "Master"
            ]
            if len(master_tag) > 0:
                master_tag = master_tag[0]
                track_meta["ai_tag"] = master_tag["what"]
                track_meta["ai_tag_confidence"] = master_tag["confidence"]

            if tag is not None:
                human_tag = tag["what"]
                human_tag_confidence = tag["confidence"]
            human_tags = [
                (t.get("what"), t["confidence"])
                for t in tags
                if t.get("automatic", False) != True
            ]

            start = None
            end = None

            prev_frame = None
            regions = {}
            for i, r in enumerate(track_meta.get("positions")):
                if isinstance(r, list):
                    region = Region.region_from_array(r[1])
                    if region.frame_number is None:
                        if i == 0:
                            frame_number = round(r[0] * FPS)
                            region.frame_number = frame_number
                        else:
                            region.frame_number = prev_frame + 1
                else:
                    region = Region.region_from_json(r)
                if region.frame_number is None:
                    if "frameTime" in r:
                        if i == 0:
                            region.frame_number = round(r["frameTime"] * 9)
                        else:
                            region.frame_number = prev_frame + 1
                prev_frame = region.frame_number
                region.frame_number = region.frame_number
                assert region.frame_number >= 0
                regions[region.frame_number] = region
                if start is None:
                    start = region.frame_number
                end = region.frame_number

            fp_meta = track_meta.get("fp_model_predictions")
            fp_frames = None
            if fp_meta is not None:
                fp_frames = []
                for pred in fp_meta.get("predictions", []):
                    scores = pred["prediction"]
                    best_arg = np.argmax(scores)
                    confidence = scores[best_arg]
                    if best_arg == fp_index and confidence > 75:
                        frame_i = pred["frames"]
                        if isinstance(frame_i, int):
                            fp_frames.append(frame_i)
                        else:
                            fp_frames.append(frame_i[0])
            header = TrackHeader(
                clip_id=clip_header.clip_id,
                track_id=int(track_meta["id"]),
                label=human_tag,
                num_frames=len(regions),
                regions=regions,
                start_frame=start,
                confidence=human_tag_confidence,
                human_tags=human_tags,
                source_file=self.file,
                mega_missed_regions=track_meta.get("mega_missed_regions"),
                station_id=clip_header.station_id,
                fp_frames=fp_frames,
                start_time=clip_header.rec_time + timedelta(seconds=start / FPS),
                # frame_temp_median=frame_temp_median,
            )
            clip_header.tracks.append(header)
        return clip_header

    def get_id(self):
        return self.meta_data_file

    def get_clip_meta(self, tag_precedence):
        return self.get_clip_tracks(tag_precedence)
        #
        # clip_meta = {}
        # clip_meta["clip_id"] = r_id
        # stationId = metadata.get("stationId", 0)
        # clip_meta["station_id"] = stationId
        # clip_meta["crop_rectangle"] = np.uint8(clip.crop_rectangle.to_ltrb())



import h5py
import numpy as np
from config.config import Config
import os

gzip_compression = {"compression": "gzip"}


def copy(old_db, new_db):
    print("copying db")

    f = h5py.File(old_db, "r")
    copy = h5py.File(new_db, "w")

    clips = f["clips"]
    copy_clips = copy.create_group("clips")
    for clip_id in clips:
        print("Copying", clip_id)
        og_frames = {}
        copy_group(clips, copy_clips, clip_id, og_frames, True)
        frame_keys = list(og_frames.keys())
        frame_keys.sort()
        copy_og = copy_clips[clip_id].create_group("original_frames")
        for k, v in og_frames.items():
            copy_frame = copy_og.create_dataset(
                str(k), shape=v.shape, chunks=v.shape, **gzip_compression, dtype=v.dtype
            )
            copy_frame[:] = v
        # break
    copy.close()
    f.close()


def copy_group(f, copy, key, og_frames, upper=False):
    copy_g = copy.create_group(key)
    group = f[key]
    for g in group:
        if g == "original" or (upper and g == "frames"):
            start_frame = 0
            if g == "original":
                start_frame = group.attrs.get("start_frame")
            f_data = group[g]
            for frame_num in f_data:
                if frame_num not in og_frames:
                    og_frames[int(frame_num) + start_frame] = f_data[frame_num]
            continue
        # print("D or G",g,isinstance(group[g], h5py.Dataset), isinstance(group[g], h5py.Group))
        if isinstance(group[g], h5py.Group):
            copy_group(group, copy_g, g, og_frames)
        else:
            # print("create new dataset", g)
            ds = group[g]
            # print(ds.shape, ds.chunks, ds.dtype, ds.compression)
            copy_ds = copy_g.create_dataset(
                g, shape=ds.shape, chunks=ds.chunks, **gzip_compression, dtype=ds.dtype
            )
            copy_ds[:] = ds

    for k, v in group.attrs.items():
        # if  isinstance(v, np.ndarray):
        #     print("numpy array", k)
        # print("key is",k,v)
        copy_g.attrs[k] = v


def main():
    config = Config.load_from_file(None)
    db_file = os.path.join(config.tracks_folder, "dataset.hdf5")
    copy(db_file, os.path.join(config.tracks_folder, "copied.hdf5"))


if __name__ == "__main__":
    main()



import time
import os
from track.clip import Clip
from track.cliptrackextractor import ClipTrackExtractor
from config.config import Config
from ml_tools.previewer import PREVIEW_TRACKING


class TestTrackingSpeed:
    CPTV_FILE_NO_BACKGROUND = "clips/hedgehog.cptv"
    CPTV_FILE_BACKGROUND = "clips/possum.cptv"

    MAX_FRAME_MS = 40

    def test_tracking_speed(self):
        config = Config.get_defaults()
        dir_name = os.path.dirname(os.path.realpath(__file__))
        file_name = os.path.join(dir_name, TestTrackingSpeed.CPTV_FILE_NO_BACKGROUND)
        print("Tracking cptv with no background ", file_name)
        track_extractor = ClipTrackExtractor(
            config.tracking,
            config.use_opt_flow or config.classify.preview == PREVIEW_TRACKING,
            cache_to_disk=False,
            verbose=config.verbose,
        )
        start = time.time()
        clip = Clip(config.tracking["thermal"], file_name)
        track_extractor.parse_clip(clip)
        ms_per_frame = (
            (time.time() - start) * 1000 / max(1, len(clip.frame_buffer.frames))
        )
        print("Took {:.1f}ms per frame".format(ms_per_frame))
        assert ms_per_frame < TestTrackingSpeed.MAX_FRAME_MS

        file_name = os.path.join(dir_name, TestTrackingSpeed.CPTV_FILE_BACKGROUND)
        print("Tracking cptv with background ", file_name)
        start = time.time()
        clip = Clip(config.tracking["thermal"], file_name)
        track_extractor.parse_clip(clip)
        ms_per_frame = (
            (time.time() - start) * 1000 / max(1, len(clip.frame_buffer.frames))
        )
        print("Took {:.1f}ms per frame".format(ms_per_frame))
        assert ms_per_frame < TestTrackingSpeed.MAX_FRAME_MS
